



def test_sparse_example_report_qpublic():
    """
    Making a meta-clonotype report from a 
    scipy.sparse TCRdist matrix.
    """
    import numpy as np
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.public import _neighbors_sparse_fixed_radius, _neighbors_sparse_variable_radius
    from tcrdist.summarize import test_for_subsets
    from tcrdist.tree import _default_sampler_olga
    from tcrdist.public import make_motif_logo_from_index

    df = pd.read_csv("dash.csv").query('epitope == "PA"')
    tr = TCRrep(cell_df = df,               #(2)
                organism = 'mouse', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = False)
        # When setting the radius to 50, the sparse matrix 
        # will convert any value > 50 to 0. True zeros are 
        # repressented as -1.
    radius = 50
    tr.cpus = 1
        # Notice that we called .compute_sparse_rect_distances instead of .compute_distances
    tr.compute_sparse_rect_distances(df = tr.clone_df, radius = radius)

        # There are two functions for finding neighbors from a sparse TCRdist matrix. 
        # For 1 fixed radius: _neighbors_sparse_fixed_radius()
        # For a radius per row: _neighbors_sparse_variable_radius()
    tr.clone_df['radius'] = 12 
    tr.clone_df['neighbors'] = \
        _neighbors_sparse_variable_radius(
            csrmat = tr.rw_beta, 
            #radius = 12)
            radius_list = tr.clone_df['radius'].to_list())

        # <K_neighbors>the number of neighbors per TCR
    tr.clone_df['K_neighbors'] = tr.clone_df['neighbors'].apply(lambda x: len(x))
        # <nsubject> the number of subject (nsubject) neighboring the TCR (
    tr.clone_df['nsubject'] = tr.clone_df['neighbors'].apply(lambda x: len(tr.clone_df['subject'][x].unique()))
        # nsubject > 1 implies quasi-publicity)
    tr.clone_df['qpublic'] = tr.clone_df['nsubject'].apply(lambda x: x >1 )
    
        # e.g., For the report, focus on TCRs with more than 5 neighboring subjects 
    quasi_public_df = tr.clone_df.query('nsubject > 5').copy().\
        sort_values('nsubject', ascending = False)
        # test_for_subsets()> allows us to remove TCRs with identical neighborhoods
    quasi_public_df['unique']  = test_for_subsets(quasi_public_df['neighbors'])
    quasi_public_df = quasi_public_df[quasi_public_df['unique'] == 1].copy()
        # declare a sampler for generating a backgrond comparison
    ts = _default_sampler_olga(organism = 'mouse', chain = 'beta')()

        # make a background-subtracted logo <svg> and raw log <svg_raw> for each TCR
    svgs = list()
    svgs_raw = list()
    for i,r in quasi_public_df.iterrows():
        svg, svg_raw  = make_motif_logo_from_index(tcrsampler = ts,
                                                   ind = r['neighbors'],
                                                   centroid = r['cdr3_b_aa'],
                                                   clone_df = tr.clone_df,
                                                   cdr3_name = 'cdr3_b_aa',
                                                   v_name = 'v_b_gene',
                                                   gene_names = ['v_b_gene','j_b_gene'])
        svgs.append(svg)
        svgs_raw.append(svg_raw)

        # Output a html report
    output_html_name = 'quasi_public_df_report.html'
    quasi_public_df['svg'] = svgs
    quasi_public_df['svg_raw'] = svgs_raw
        # Specific columns to include in the report
    labels = [  'cdr3_b_aa', 
                'v_b_gene',
                'j_b_gene',
                'radius', 
                'K_neighbors',
                'nsubject']

    def shrink(s):
        """reduce size of svg graphic"""
        s = s.replace('height="100%"', 'height="20%"')
        s = s.replace('width="100%"', 'width="20%"')
        return s

    with open(output_html_name, 'w') as output_handle:
        for i,r in quasi_public_df.iterrows():
            #import pdb; pdb.set_trace()
            svg, svg_raw = r['svg'],r['svg_raw']
            output_handle.write("<br></br>")
            output_handle.write(shrink(svg))
            output_handle.write(shrink(svg_raw))
            output_handle.write("<br></br>")
            output_handle.write(pd.DataFrame(r[labels]).transpose().to_html())
            output_handle.write("<br></br>")


"""
These are unit tests
"""
import numpy as np
import re

flatten = lambda l: [item for sublist in l for item in sublist]

"""
These are not tests per say but useful illustrations
"""
def allele_01(genename):
	"""
	>>> allele_01('TRBV19*02')
	'TRBV19*01'
	"""
	g,a = re.match(pattern = '(.*)([0-9])$', string= genename).groups()
	return(f"{g}1")


def test_olga_sample_beta():
	np.random.seed(1)
	from tcrdist.pgen import OlgaModel
	olga_model_beta = OlgaModel(recomb_type="VDJ", chain_folder = "human_T_beta")
	result = olga_model_beta.gen_cdr3s(V = 'TRBV20-1*01', J = 'TRBJ1-2*01', n = 5)
	assert isinstance(result, list)
	assert len(result) == 5
	assert result == ['CSARQGLANYGYTF','CSARPSRGQDGYTF','CSARDQRTGQDGYTF','CSARDVSSSGGYYGYTF','CSAPEPLTSGRACNGYTF']

def test_olga_sample_alpha():
	np.random.seed(1)
	from tcrdist.pgen import OlgaModel
	olga_model_alpha = OlgaModel(recomb_type="VJ", chain_folder = "human_T_alpha")
	result = olga_model_alpha.gen_cdr3s(V = 'TRAV19*01', J = 'TRAJ37*01', n = 5)
	assert isinstance(result, list)
	assert len(result) == 5
	assert result == ['CALSEAPGNTGKLIF','CAPPSGNTGKLIF','CALAGNTGKLIF','CAQDNTGKLIF','CALRNTGKLIF']

def test_olga_sample_alphas_for_a_human_repertoire():
	import re
	import pandas as pd
	from tcrdist.repertoire import TCRrep
	import palmotif

	from tcrdist.pgen import OlgaModel
	olga_model_alpha = OlgaModel(recomb_type="VJ", chain_folder = "human_T_alpha")

	from tcrdist.pgen import OlgaModel
	olga_model_beta  = OlgaModel(recomb_type="VDJ", chain_folder = "human_T_beta")


	df = pd.read_csv("dash_human.csv")
	tr = TCRrep(cell_df = df, 
	            organism = 'human', 
	            chains = ['alpha','beta'], 
	            db_file = 'alphabeta_gammadelta_db.tsv')

	rb = [olga_model_beta.gen_cdr3s(V = allele_01(r['v_b_gene']), J = allele_01(r['j_b_gene']), n = 1) for _,r in tr.clone_df[['v_b_gene', 'j_b_gene']].iterrows()]
	ra = [olga_model_alpha.gen_cdr3s(V = allele_01(r['v_a_gene']), J = allele_01(r['j_a_gene']), n = 1) for _,r in tr.clone_df[['v_a_gene', 'j_a_gene']].iterrows()]
	
	# assert that we covered 95% of dash_human wiht our olga sampler 
	#assert len([x for x in flatten(rb) if x is not None]) / len(flatten(rb)) > .95
	#assert len([x for x in flatten(ra) if x is not None]) / len(flatten(ra)) > .95


def test_olga_sample_alphas_for_a_large_repertoire():
	import re
	import pandas as pd
	from tcrdist.repertoire import TCRrep
	import palmotif

	from tcrdist.pgen import OlgaModel
	olga_model_beta_mouse = OlgaModel(recomb_type="VDJ", chain_folder = "mouse_T_beta")


	df = pd.read_csv("dash.csv")
	tr = TCRrep(cell_df = df, 
	            organism = 'mouse', 
	            chains = ['beta'], 
	            db_file = 'alphabeta_gammadelta_db.tsv')

	rb = [olga_model_beta_mouse.gen_cdr3s(V = allele_01(r['v_b_gene']), J = allele_01(r['j_b_gene']), n = 1) for _,r in tr.clone_df[['v_b_gene', 'j_b_gene']].iterrows()]
	# assert that we covered 95% of dash_human wiht our olga sampler 
	# assert len([x for x in flatten(rb) if x is not None]) / len(flatten(rb)) > .95
	


def test_olga_sample():
	np.random.seed(310)
	from tcrdist.pgen import OlgaModel
	olga_model_beta = OlgaModel(recomb_type="VDJ", chain_folder = "human_T_beta")
	result = olga_model_beta.gen_cdr3(V = 'TRBV20-1*01', J = 'TRBJ1-2*01')
	# NOTE: seed is set, so we expect standard result
	#NOTE: .gen_cdr3() returns the full output tuple
	expected = ('TGCAGTGCTAGAGTAAGGGAAGCGGGAAGGACCTACACCTTC',
				 'CSARVREAGRTYTF',
				 29,
				 1,
				 {'V': 29,
				  'D': 2,
				  'J': 1,
				  'delV': 5,
				  'delJ': 15,
				  'delDl': 10,
				  'delDr': 7,
				  'insVD': 7,
				  'insDJ': 6})
	assert result == expected
	
	# NOTE: .gen_cdr3s() returns a list of CDR3s (amino acid only)
	np.random.seed(310)
	result = olga_model_beta.gen_cdr3s(V = 'TRBV20-1*01', J = 'TRBJ1-2*01', n = 4)
	expected = ['CSARVREAGRTYTF', 'CSAVPPGLPNYGYTF', 'CSARGPSQGYVRGLYGYTF', 'CSAQGLAGYGYTF']
	assert result == expected



def motif_creation_human_betas():
	import re
	import pandas as pd
	from tcrdist.repertoire import TCRrep
	import palmotif

	from tcrdist.pgen import OlgaModel
	oma = OlgaModel(recomb_type="VJ", chain_folder = "human_T_alpha")

	from tcrdist.pgen import OlgaModel
	omb = OlgaModel(recomb_type="VDJ", chain_folder = "human_T_beta")


	df = pd.read_csv("dash_human.csv")
	tr = TCRrep(cell_df = df, 
	            organism = 'human', 
	            chains = ['alpha','beta'], 
	            db_file = 'alphabeta_gammadelta_db.tsv')

	from tcrdist.adpt_funcs import get_basic_centroids
	get_basic_centroids(tr, max_dist = 75)
	with open("test_3.svg", 'w') as oh:
		oh.write('<body>')
		for i,r in tr.centroids_df.iterrows():
			if len(r['neighbors']) < 5:	
				break
			seqs = tr.clone_df.iloc[r['neighbors'],]['cdr3_b_aa'].to_list()
			gene_usages = tr.clone_df.iloc[r['neighbors'],][['v_b_gene', 'j_b_gene']].value_counts().reset_index().to_dict('split')['data']
			depth = 3

			refs = flatten([omb.gen_cdr3s(allele_01(v),allele_01(j),i*depth) for v,j,i in combos_alpha])
			refs = [x for x in refs if x is not None]
			
			matrix, stats = palmotif.compute_pal_motif(seqs = seqs , refs = refs, centroid = r['cdr3_b_aa'])
			matrix_raw, _ = palmotif.compute_pal_motif(seqs = seqs , centroid = r['cdr3_b_aa'])
			refs.append(r['cdr3_b_aa'])
			matrix_bkgd, _ = palmotif.compute_pal_motif(seqs = refs, centroid = r['cdr3_b_aa'])
			
			svgs = [palmotif.svg_logo(matrix, 'test.svg', return_str = True), 
					palmotif.svg_logo(matrix_raw, 'test.svg', return_str = True),
					palmotif.svg_logo(matrix_bkgd, 'test.svg', return_str = True)]

			[oh.write(f"{s}<div></div>\n") for s in svgs]
			oh.write('<div></div>')
			oh.write(str(r))
			oh.write('<div></div>')
		
		oh.write('</body>')



def sim_all_cdr3_gen():
	import itertools
	def expand_grid(dct):
	    rows = itertools.product(*dct.values())
	    return pd.DataFrame.from_records(rows, columns=dct.keys())


	from tcrdist.pgen import OlgaModel
	omb = OlgaModel(recomb_type="VDJ", chain_folder = "human_T_beta")
	all_possible_beta = expand_grid(
	    {'V': omb.pgen_model.V_allele_names,
	     'J': omb.pgen_model.J_allele_names}
	)

	find_nones = list()
	results = list()
	for i,r in all_possible_beta.iterrows():
		e = omb.gen_cdr3s(V = r['V'], J = r['J'], n = 3)
		results.append(e)
		if e is None:
			find_nones.append( [r['V'],r['J'], e])
		print((r['V'],r['J'],e))



	from tcrdist.pgen import OlgaModel
	oma = OlgaModel(recomb_type="VJ", chain_folder = "human_T_alpha")
	all_possible_alpha = expand_grid(
	    {'V': oma.pgen_model.V_allele_names,
	     'J': oma.pgen_model.J_allele_names}
	)

	find_nones = list()
	results = list()
	for i,r in all_possible_alpha.iterrows():
		e = oma.gen_cdr3(V = r['V'], J = r['J'])
		results.append([r['V'],r['J'], e])
		if e is None:
			find_nones.append( [r['V'],r['J'], e])
		print((r['V'],r['J'],e))

	# Things we can't find:
	df = pd.DataFrame(results, columns = ['v','j','r'])
	df[df['r'].isna()][['v']].value_counts()
	df[df['r'].isna()][['j']].value_counts()












def test_neighbors_and_publicity_directly():
    """
    Instead of enforcing a fixed radius, 
    use a radius specific to each
    centroid, specified in an additional 
    column.
    """
    import os
    import pandas as pd
    import numpy as np
    from tcrdist.repertoire import TCRrep
    from tcrdist.public import TCRpublic  
    fn = os.path.join('tcrdist','data','covid19',
        'mira_epitope_55_524_ALRKVPTDNYITTY_KVPTDNYITTY.tcrdist3.radius.csv')
    df = pd.read_csv(fn)
    tr = TCRrep(cell_df = df[['cohort','subject','v_b_gene', 'j_b_gene','cdr3_b_aa', 'radius']], 
                organism = "human", 
                chains = ["beta"])
    
    # NEIGHBORS
    from tcrdist.public import _neighbors_fixed_radius
    from tcrdist.public import _neighbors_variable_radius
    # returns lists of lists of all neighbors at fixed of variable radii
    _neighbors_fixed_radius(pwmat = tr.pw_beta, radius = 18)
    _neighbors_variable_radius(pwmat = tr.pw_beta, radius_list = tr.clone_df.radius)

    # returns the number (K) neighbors at fixed or vriable radii
    from tcrdist.public import _K_neighbors_fixed_radius
    from tcrdist.public import _K_neighbors_variable_radius
    _K_neighbors_fixed_radius(pwmat = tr.pw_beta, radius = 18)
    _K_neighbors_variable_radius(pwmat = tr.pw_beta, radius_list = tr.clone_df.radius)

    # First find neighbors by your favorite method 
    tr.clone_df['neighbors'] = _neighbors_variable_radius(
        pwmat = tr.pw_beta, 
        radius_list = tr.clone_df.radius)
    # Once neighbors are added to a clone_df you can easily determine publicity. 
    tr.clone_df['nsubject']   = tr.clone_df['neighbors'].\
        apply(lambda x: tr.clone_df['subject'].iloc[x].nunique())
    tr.clone_df['qpublic']   = tr.clone_df['nsubject'].\
        apply(lambda x: x > 1)



"""
Simple set of functions for summarizing over a group
"""

import pandas as pd
import numpy as np
import re


def filter_is(df, col, val):
    return df[df[col] == val]

def filter_in(df, col, vals):
    ind = [np.any(x in vals) for x in df[col]]
    return df[ind]

def filter_gt(df, col, val):
    return df[df[col] > val]

def filter_lt(df, col, val):
    return df[df[col] < val]


def is_subset(set1, set2):
    """
    Return True, if all members of set2 are contained in set1. 
    i.e, set2 is a subset of set1. See example for clarity:
    Example
    -------
    >>> is_subset(set(["A","B"]), set(["A"]))
    True
    >>> is_subset(set(["A","B"]), set(["A","C"]))
    False
    """
    return set(set2)-set(set1) == set()


def is_almost_subset(set1, set2, min_set_diff = 2):
    """
    Return True, if no more than min_set_diff members of set2 
    are contained in set1. 
    i.e, set2 is almost a subset of set1. See example for clarity:
    
    Example
    -------
    >>>  is_almost_subset(set(["A","B","C","D"]), set(["A", "K"]), 2)
    True
    >>> is_almost_subset(set(["A","B","C","D"]), set(["A", "K"]), 1)
    False

    """
    return len(set(set2)-set(set1)) < min_set_diff


def test_for_subsets(list_of_sets):
    """

    test_for_subsets (formerly known as turtles_all_the_way_down)

    For a ranked list of sets, return a vector where 1 
    indicated the set is not a subset of any of sets
    that come before it in the list. 

    This is useful for eliminating clusters (sets) of TCRs
    which are smaller than a higher ranked and larger set 
    that contains all its members. See example for clarity:

    Example 
    -------
    >>> test_for_subsets([["A","B","C"], ["A","C","D"], ["A","D"], ["B","E"],["B","C"]])
    [1, 1, 0, 1, 0] 
    >>> test_for_subsets([ [1,2,3], [1,3,4], [1,4], [2,5],[2,3]])
    [1, 1, 0, 1, 0]
    """
    tracker = [1]
    if isinstance(list_of_sets, pd.Series):
        list_of_sets = list_of_sets.to_list()
    checked_sets = [list_of_sets[0]]
    for s in list_of_sets[1:]:
        if np.any([is_subset(cs, s) for cs in checked_sets]):
            tracker.append(0)
        else: 
            tracker.append(1)
            checked_sets.append(s)
    assert len(tracker) == len(list_of_sets)
    return tracker


def test_for_almost_subsets(list_of_sets, thr = 3):
    """

    test_for_subsets (formerly known as turtles_all_the_way_down)

    For a ranked list of sets, return a vector where 1 
    indicated the set is not a subset of any of sets
    that come before it in the list. 

    This is useful for eliminating clusters (sets) of TCRs
    which are smaller than a higher ranked and larger set 
    that contains all its members. See example for clarity:

    Example 
    -------
    >>> test_for_almost_subsets([["A","B","C"], ["A","C","D"], ["A","D"], ["B","E"],["B","C"]], 1)
    [1, 1, 0, 1, 0] 
    >>> test_for_almost_subsets([ [1,2,3], [1,3,4], [1,4], [2,5],[2,3]], 1)
    [1, 1, 0, 1, 0]
    """
    tracker = [1]
    if isinstance(list_of_sets, pd.Series):
        list_of_sets = list_of_sets.to_list()
    checked_sets = [list_of_sets[0]]
    for s in list_of_sets[1:]:
        if np.any([is_almost_subset(cs, s, thr) for cs in checked_sets]):
            tracker.append(0)
        else: 
            tracker.append(1)
            checked_sets.append(s)
    assert len(tracker) == len(list_of_sets)
    return tracker



def _dist_summ(data, precision = 1, scientific = True):
    """
    Summarise distribution [as min,q1,median,q3, max]
    
    Parameters
    ----------
    data : list
        List of numeric data
    precision : int
        How many integers precision in scientific notation = 1, 
    scientific : bool
        Default is True, to return result in scientific notation 

    Examples
    --------
    >>> _dist_summ([1,2,3,4,5])
    ['1.e+00', '2.e+00', '3.e+00', '4.e+00', '5.e+00']
    
    _dist_summ([1,2,3,4,5], scientific=False)
    [1, 2.0, 3.0, 4.0, 5]

    """
    dmin = np.min(data)
    dQ1 = np.percentile(data, q = 25, interpolation = 'midpoint') 
    dmedian = np.median(data)
    dQ3 = np.percentile(data, q = 75, interpolation = 'midpoint') 
    dmax = np.max(data)
    r = [dmin, dQ1, dmedian, dQ3, dmax]
    if scientific:
        return [np.format_float_scientific(s, precision = precision) for s in r]
    else:
        return r 



def _select(df, iloc_rows, col = 'cdr3_b_aa'):
    return df.iloc[iloc_rows,][col].to_list()

def _summ(df, indices, column = None , f=None, fdf = None, **kwargs):
    """
    _summ implements a split, apply some function, combine result routine. 
    
    Parameters
    ----------
    f : callable
        a function callable on a list of series
    fdf : callable
        a function callable on a dataframe
    df : pd.DataFrame
        DataFrame
    indices : list
        list of lists containing integers corresponding to the iloc rows of a the < df >
    column : str or None
        column name, should be None if using a fdf
        
    Returns
    -------
    summary : list of identical lenght to indices

    Examples
    --------
    >>> from tcrdist.summarize import _summ, _occurs_N_str, _top_N_str
    >>> df = pd.DataFrame({'catvar':["a","b","b","c"], "numvar":[10,1,100,3]})
    >>> _summ(df, indices = [[0,1], [2,3]], column = 'numvar', f = np.median)
    [5.5, 51.5]
    >>> _summ(df, indices = [[0,1], [2,3]], column = 'catvar', f = _occurs_N_str, N = 2)
    ['b (50.0%), a (50.0%)', 'c (50.0%), b (50.0%)']
    >>> _summ(df, indices = [[0,1], [2,3]], column = 'catvar', fdf = _top_N_str, **{'col': 'catvar', 'count_col': 'numvar','N':2})
    ['a (90.9%), b (9.1%)', 'b (97.1%), c (2.9%)']
    """
    summary = list()
    for ind in indices:
        if f is not None:
            if isinstance(df.iloc[ind, ][column], pd.Series):
                selection = df.iloc[ind, ][column].to_list()
            else:
                selection = df.iloc[ind, ][column]
            summary.append(f(selection, **kwargs))
        elif fdf is not None:
            selection = df.iloc[ind, ]
            summary.append(fdf(selection, **kwargs))
        else:
            raise(ValueError("No function (f) or function on a DataFrame (fdf) were supplied\n"))
    assert len(summary) == len(indices)
    return summary

def _occurs_N_str(m, N):
    """
    Return occurances in a pd.Series as a string 

    Example
    -------
    >>> _occurs_N_str(["a","b","b","c"], 1)
    'b (50.0%)'    
    >>> _occurs_N_str(["a","b","b","c"], 2)
    'b (50.0%), c (25.0%)' 
    >>> _occurs_N_str(["a","b","b","c"], 3)
    'b (50.0%), c (25.0%), a (25.0%)'
    """
    if isinstance(m, pd.Series):
        gby = m.value_counts()
    else:
        m = pd.Series(m)
        gby = m.value_counts()
    gby = 100 * gby / gby.sum()
    gby = gby.sort_values(ascending=False)
    out = ', '.join(['%s (%2.1f%%)' % (idx, v) for idx,v in gby.items()][:N])
    return out


def _top_N_str(m, col, count_col, N):
    """
    Example
    -------
    >>> df = pd.DataFrame({'catvar':["a","b","b","c"], "numvar":[10,1,100,3]})
    >>> _top_N_str(df, col = 'catvar', count_col ='numvar', N=2)
    'b (88.6%), a (8.8%)'
    """
    gby = m.groupby(col)[count_col].agg('sum')
    gby = 100 * gby / gby.sum()
    gby = gby.sort_values(ascending=False)
    out = ', '.join(['%s (%2.1f%%)' % (idx, v) for idx,v in gby.items()][:N])
    return out


def _extract_percentage(s, key):
    """
    extractor for pattern '%s (%2.1f%%)', see examples for clarity

    Parameter
    ---------
    s : str
        string pattern '%s (%2.1f%%)','%s (%2.1f%%)','%s (%2.1f%%)'
    k : str
        key for the percentage you want to extract

    Returns 
    -------
    tuple (str, float)
    
    Examples 
    --------
    >>> _extract_percentage('naive_CD8 (100.0%)', 'naive_CD8')
    ('naive_CD8', '100.0')

    >>> _extract_percentage('naive_CD8 (100.0%)', 'PBMC')
    ('PMBC', 0.0)

    >>> _extract_percentage('naive_CD8 (94.1%), PBMC (5.9%)', 'PBMC')
    ('PBMC', '5.9')
    """
    ls = s.split(",")
    try: 
        rs = [re.search(pattern = '([A-Za-z0-9_]+) [(]([0-9]+[\.][0-9])%[)]', string = s) for s in ls]
        rgs =  [reg.groups() for reg in rs]
        return (key, {k:v for k,v in rgs}[key])
    except:
        return key, 0.0



def member_summ(res_df, clone_df, key_col = 'neighbors_i', count_col='count', addl_cols=[], addl_n=1):
    """Return additional summary info about each result (row)) based on the members of the cluster.

    This is helpful for preparing strings to add to the tooltip in hierdiff.plot_hclust_props.
    
    Parameters
    ----------
    res_df : pd.DataFrame [nclusters x result cols]
        Returned from neighborhood_diff or hcluster_diff
    clone_df : pd.DataFrame [nclones x metadata]
        Contains metadata for each clone.
    key_col : str
        Column in res_df that specifies the iloc of members in the clone_df
    count_col : str
        Column in clone_df that specifies counts.
        Default none assumes count of 1 cell for each row.
    addl_cols : list
        Columns to summarize
    addl_n : int
        Number of top N clones to include in the summary of
        each cluster.

    Returns
    -------
    summ : pd.DataFrame [nclusters x summary columns]
        Columns that can be joined with res_df


    Example
    -------
    summ_df = member_summ(res_df, clone_df)
    res_df = res_df.join(summ_df, how='left')"""
    def _top_N_str(m, col, count_col, N):
        gby = m.groupby(col)[count_col].agg('sum')
        gby = 100 * gby / gby.sum()
        gby = gby.sort_values(ascending=False)
        out = ', '.join(['%s (%2.1f%%)' % (idx, v) for idx,v in gby.items()][:N])
        return out
    
    split = []
    for resi, res_row in res_df.iterrows():
        m = clone_df.iloc[res_row[key_col]]

        mode_i = m[count_col].idxmax()
        summ = {}
        for c in [c for c in clone_df.columns if 'cdr3' in c]:
            summ[c] = _top_N_str(m, c, count_col, 1)
        for c in [c for c in clone_df.columns if 'gene' in c]:
            summ[c] = _top_N_str(m, c, count_col, 3)

        x_val_cols = [c for c in res_df.columns if 'x_val_' in c]
        x_freq_cols = [c for c in res_df.columns if 'x_freq_' in c]
        
        for label_col, freq_col in zip(x_val_cols, x_freq_cols):
            summ[res_row[label_col]] = np.round(res_row[freq_col], 3)

        for c in [c for c in addl_cols]:
            summ[c] = _top_N_str(m, c, count_col, addl_n)
        summ = pd.Series(summ, name=resi)
        split.append(summ)
    summ = pd.DataFrame(split)
    return summ


import numpy as np
import pandas as pd
from scipy import sparse
from matplotlib import pyplot as plt
import matplotlib as mpl

__all__ = ['distance_ecdf',
           'make_ecdf_step',
           'plot_ecdf']

def distance_ecdf(pwrect, thresholds=None, weights=None, pseudo_count=0, skip_diag=False, absolute_weight = False):
    """Computes the empirical cumulative distribution function (ECDF) for
    each TCR in a set of target TCRs [rows of pwrect] as the proportion
    of reference TCRs [columns of pwrect] within a distance radius less
    than or equal to a threhold d_i, over a range of
    D = [d_1, d_2, ..., d_i]. The distances between pairs of TCRs in the
    target and reference set are contained in the elements of pwrect.

    Optionally, relative weights can be supplied for each reference TCR.
    These can be TCR counts or other weights and the ECDF will still
    be a probability on [0, 1].

    Parameters
    ----------
    pwrect : np.ndarray or scipy.sparse.csr_matrix, (clone_df.shape[0], n_ref)
        Matrix of pairwise distances among TCRs (can be rectangular). Sparse matrix may encode
        true zeros (i.e. identical TCRs) as -1 or 0, without having effect on the ECDF.
    thresholds : np.ndarray
        Vector of thresholds at which the ECDF should be evaluated.
        By default will use all unique values in pwrect. For sparse arrays the thresholds
        should not exceed the largest distance encoded.
    weights : np.ndarray or list, (clone_df.shape[0], )
        Relative weight of each TCR in the reference (column dimension of pwrect)
    pseudo_count : int
        Added to the numerator and denominator at each threshold
        to avoid zero. Useful if end goal is a log-scale plot.
    skip_diag : bool
        Skip counting the diagonal for computing ECDF of seqs against same seqs.
    absolute_weight : bool
        if True, denominator is number of total sequences in pwrect.shape[2] rather than sum of the weights.
        
    Returns
    -------
    thresholds : vector, thresholds.shape[0]
    ecdf : vector, thresholds.shape[0]"""
    if weights is None:
        weights = np.ones(pwrect.shape[1])
    else:
        weights = np.asarray(weights)

    if thresholds is None:
        if sparse.issparse(pwrect):
            thresholds = np.unique(pwrect.data)
        else:
            thresholds = np.unique(pwrect[:])
    else:
        thresholds = np.asarray(thresholds)

    """Vectorized and faster, using broadcasting for the <= expression"""
    """ecdf = np.sum((pwrect[:, :, None] <= thresholds[None, None, :]) * \
            weights[None, :, None], axis=1) / np.sum(weights)"""
    
    """Decided not to vectorize in 3D because it can create an array that's
    too big for memory"""
    ecdf = np.zeros((pwrect.shape[0], thresholds.shape[0]))
    sum_weights = np.sum(weights)
    if absolute_weight:
        denom = pwrect.shape[1]

    for i in range(pwrect.shape[0]):
        if sparse.issparse(pwrect):
            """By using the row.data attribute we only sum over distances
            that were stored: large distances that were discarded from the sparse
            representation (which are sometimes thought of as zeros) are simply
            not present and therefore never counted as below threshold[i]"""
            row = pwrect[i, :]
            numer = np.sum((row.data[:, None] <= thresholds[None, :]) * weights[row.indices, None], axis=0)
        else:
            row = np.reshape(pwrect[i, :], (pwrect.shape[1], 1))
            numer = np.sum((row <= thresholds[None, :]) * weights[:, None], axis=0)
        if absolute_weight:
            denom = pwrect.shape[1]
        else:
            denom = sum_weights
        if skip_diag:
            numer = numer - weights[i]
            denom = denom - weights[i]
        ecdf[i, :] = (numer + pseudo_count) / (denom + pseudo_count)
    return thresholds, ecdf

def make_ecdf_step(thresholds, ecdf, add_mnx=False, add_mny=False, add_mnmn=False,enforce_mn=False , mn=(0, 0), xjitter=0):
    """Create stepped vector for plotting an ECDF,
    since the ECDF should naturally have discrete steps
    but will not unless they are explictly added prior
    to plotting.

    Takes outputs from distance_ecdf function, but
    is general for creating stepped vectors for plotting.

    Parameters
    ----------
    thresholds : np.ndarray [n_thresholds, ]
        Vector of thresholds
        (typically the x-axis of the plot)
    ecdf : np.ndarray [n_thresholds, ]
        Vector of increasing probabilities
        (typically the y-axis of the plot)
    add_mnmn : bool
        Will add a step to (x=mn[0], y=mn[1]) to complete
        the ECDF.

    Returns
    -------
    x, y : np.ndarray
        Vectors for plotting"""
    y = np.asarray(ecdf)
    t = np.asarray(thresholds)

    if enforce_mn:
        t[t<mn[0]] = mn[0]
        y[y<mn[1]] = mn[1]

    if add_mnx:
        t = np.concatenate(([mn[0]], t.ravel()))
        y = np.concatenate(([y[0]], y.ravel()))
    if add_mny:
        t = np.concatenate(([t[0]], t.ravel()))
        y = np.concatenate(([mn[1]], y.ravel()))

    if add_mnmn:
        t = np.concatenate(([mn[0]], t.ravel()))
        y = np.concatenate(([mn[1]], y.ravel()))

    t = np.concatenate(([t[0]], np.repeat(t[1:].ravel(), 2)))
    y = np.repeat(y.ravel(), 2)[:-1]

    jx = t + (np.random.rand(1) - 0.5) * xjitter
    jx[0] = t[0]
    jx[-1] = t[-1]
    return jx, y

def gmean10(vec, axis=0):
    """Geometric mean which may be useful for
    summarizing many ECDF functions"""
    return 10 ** (np.mean(np.log10(vec), axis=axis))


def plot_ecdf(
    thresholds,
    ecdf_mat, 
    ax = None,
    ylim = None,
    xlabel = f'Distance From Target TCR Clone',
    ylabel = f'Proportion of Reference TCRs', 
    plot_mean = True,
    min_freq = 1E-10):
    """
    A very basic ecdf plot

    Parameters
    ----------
    thresholds : np.array
        1D array tcrdistance threshold trdistance units (tdus)
    ecdf_mat : np.array
        2D array, each row repressents a TCR and the proportion of neighbors with tdus
    ax : matplotlib axes or None

    ylabel : str
        'Proportion of X TCRs'
    xlabel : str
        'Proportion of Reference TCRs'
    plot_mean : bool
        If true, plot the mean value
    """
    if ax is None:
        ax = plt.gca()
    if not ylim is None:
        ax.set_ylim(ylim)

    ax.set_yscale('log')
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)
    
    for row_i in range(ecdf_mat.shape[0]):
        make_ecdf_step(thresholds, ecdf_mat[row_i, :], add_mnx=True, enforce_mn=True, mn=(0, min_freq), xjitter=1)
        x, y = make_ecdf_step(thresholds, ecdf_mat[row_i, :])
        ax.plot(x, y, color='k', alpha=0.2)
    if plot_mean:
        x, y = make_ecdf_step(thresholds, np.mean(ecdf_mat, axis=0))
        ax.plot(x, y, color='r', alpha=1)
    return True



def _plot_manuscript_ecdfs(
    thresholds, 
    ecdf_mat, 
    ylab = 'Proportion of X TCRs', 
    cdr3_len=None, 
    min_freq=1e-6,
    cdr3_len_min=10., 
    cdr3_len_max=16.,
    low_pass = -5,
    cmap=mpl.cm.viridis_r):
    """
    _plot_manuscript_ecdfs, recreate the manuscript type ecdf

    Parameters
    ----------
    thresholds : np.array
        1D array tcrdistance threshold trdistance units (tdus)
    ecdf_mat : np.array
        2D array, each row repressents a TCR and the proportion of neighbors with tdus
    ylab : str
        'Proportion of X TCRs'
    cdr3_len : list or pd.Series
        Lengths of cdr3, must match order and dimension of rows in ecdf_mat
    min_freq : float
        Minimum frequency or proportion to display
    cdr3_len_min : float 

    cdr3_len_max : float
    
    low_pass : int
           controls threshold frequencyh to show on lower gray bar (-5) corresponds with 10^-5

    """
    # Define colormap range for CDR3 length
    low_pass_num = 10**low_pass # -5 -> 10^-5
    norm = mpl.colors.Normalize(vmin=cdr3_len_min, vmax=cdr3_len_max)
    if cdr3_len is None:
        cdr3_len = 10 * np.ones(ecdf_mat.shape[0])

    # Declare Figure <figh>
    figh = plt.figure(figsize=(11, 8))
    # Declare Grid
    gs = plt.GridSpec(nrows=2, ncols=2, width_ratios=[10, 1], height_ratios=[7, 1], hspace=0.15)
        
    # [0,1] : Assign color bar axis
    cbar = mpl.colorbar.ColorbarBase(figh.add_subplot(gs[0, 1]), cmap=cmap,
                            norm=norm,
                            orientation='vertical')
    cbar.set_label('CDR3 length')

    # [0,0] : Asign ECDF axis
    axh = figh.add_subplot(gs[0, 0], yscale='log')
    axh.set_ylabel(ylab)
    # plt.xlabel(f'Neighborhood radius\n(tcrdist units)')
    for tari in range(ecdf_mat.shape[0]):
        x, y = make_ecdf_step(thresholds, ecdf_mat[tari, :], add_mnx=True, enforce_mn=True, mn=(0, min_freq), xjitter=1)
        axh.plot(x, y, color=mpl.cm.viridis_r(norm(cdr3_len[tari]))[:3], alpha=0.4)
    plt.annotate(text=f'n={ecdf_mat.shape[0]}',
                 xy=(min_freq, 0.5),
                 xytext=(3, -3),
                 textcoords='offset points',
                 ha='left',
                 va='top')
    axh.set_xticklabels([])
    axh.set_ylim((min_freq, 0.5))
    axh.set_xlim((0, thresholds[-1]))

    # [1, 0] : Assign underplot
    axh2 = figh.add_subplot(gs[1, 0])
    x, y = make_ecdf_step(thresholds, np.mean(ecdf_mat<low_pass_num, axis=0), add_mnx=False)
    axh2.fill_between(x, np.zeros(y.shape[0]), 100*y, color='gray')
    axh2.set_ylim((0, 100))
    axh2.set_ylabel(f'% TCRs\n$ECDF < 10^{low_pass}$')
    axh2.set_xlabel(f'Neighborhood radius\n(tcrdist units)')
    axh2.set_xlim((0, thresholds[-1]))
    return figh



import logging
from collections import Counter
import os
import pandas as pd
import numpy as np
import pwseqdist as pw
from tcrdist.rep_funcs import _pw, _pws

import scipy
from scipy.cluster.hierarchy import linkage, dendrogram, fcluster
from scipy.spatial.distance import squareform

from tcrdist.repertoire import TCRrep
from tcrdist.swap_gene_name import adaptive_to_imgt

__all__ = [ 'import_adaptive',
            'ispublic',
            'simple_cluster_index',
            'cluster_index_to_df',
            'get_centroid_seq',
            'default_tcrdist_and_image']


def import_adaptive_file(   adaptive_filename, 
                            organism = "human", 
                            chain = "beta",
                            return_valid_cdr3_only = True, 
                            count = 'productive_frequency',
                            version_year= 2020,
                            sep = "\t",
                            subject = None,
                            epitope = None,
                            log = True, 
                            swap_imgt_dictionary = None,
                            additional_cols = None,
                            use_cols =  ['bio_identity', 'productive_frequency', 'templates', 'rearrangement']):
    """
    Prepare tcrdist3 input from 2020 Adaptive File containing 'bio_identity', 'productive_frequency', 'templates', and 'rearrangement'.

    Parameters 
    ----------
    adaptive_filename : str
        path to the Adaptive filename 
    version : int
        version_year
    epitope : str or None
        name of epitope if known
    subject : str or None
        If none the filename will be used as the subject
    use_as_count : str
        name of column to be used as count (could be 'productive_frequency' or 'templates')
    sep : str
        seperatore in Adaptive file
    organism : str
        'human' or 'mouse'
    chain : str
        'beta' or 'alpha'
    log : bool
        If True, write a log.
    swap_imgt_dictionary : dict or None
        If None, the default dictionary adaptive_to_imgt is used
    additional_cols : None or List
        list of any additional columns you want to keep
    use_cols : list
        ['bio_identity', 'productive_frequency', 'templates', 'rearrangement','subject'] list of columns to retain from original input file. Add 'subject' 
        if you wish to retain the subject.
    
    Returns 
    -------
    bulk_df : pd.DataFrame
    """
    try:
        bulk_df = pd.read_csv(adaptive_filename, sep= sep, usecols = use_cols )
    except ValueError as e:
        raise Exception('Bulk Adpative TCR file was missing required columns') from e


    if swap_imgt_dictionary is None:
        swap_imgt_dictionary = adaptive_to_imgt

    if log: logging.basicConfig(filename='prepbulk.log',level=logging.DEBUG, format='tcrdist3:%(asctime)s\n\t %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')

    item_names = {'alpha': ["cdr3_b_aa","v_a_gene", "j_a_gene","cdr3_b_nucseq"], 
                  'beta' : ["cdr3_b_aa","v_b_gene","j_b_gene","cdr3_b_nucseq"]}[chain]

    # Coerce strings to numeric
    bulk_df['productive_frequency'] = pd.to_numeric(bulk_df['productive_frequency'],errors='coerce')

    # Parse bio-identity
    ns= {0:"cdr3_aa", 1:"v_gene", 2:"j_gene"}
    cdr_v_j = bulk_df['bio_identity'].str.split("+", expand = True).\
        rename(columns = lambda x: ns[x])
    bulk_df[[item_names[0], 'v_gene', 'j_gene']] = cdr_v_j

    # Convert Names from Adapative to IMGT
    bulk_df[item_names[1]] = bulk_df['v_gene'].apply(lambda x : swap_imgt_dictionary[organism].get(x))
    bulk_df[item_names[2]] = bulk_df['j_gene'].apply(lambda x : swap_imgt_dictionary[organism].get(x))
    # Record Invalid Names
    invalid_v_names =  Counter(bulk_df['v_gene'][  bulk_df[item_names[1]].isna() ].to_list())
    invalid_j_names =  Counter(bulk_df['j_gene'][  bulk_df[item_names[2]].isna() ].to_list())

     # Validate CDR sequences
    bulk_df['valid_cdr3'] = bulk_df[item_names[0]].apply(lambda cdr3: _valid_cdr3(cdr3)) 
    # Count number of valid seqs
    valid = np.sum(bulk_df['valid_cdr3'])
    
    # Assign subject baesd on the < subject > argument if not already in bulk_df
    if 'subject' not in bulk_df.columns:
        if subject is None:
            bulk_df['subject'] = adaptive_filename
        else: 
            bulk_df['subject'] = subject

    # Assign a user supplied or blank epitope baesd on the < epitope > argument 
    if epitope is None:
        bulk_df['epitope'] = 'X' 
    else: 
        bulk_df['epitope'] = epitope

    if additional_cols is None:
        bulk_df = bulk_df[['subject','productive_frequency', 'templates','epitope',item_names[0],item_names[1],item_names[2],'valid_cdr3','rearrangement']].copy()
        bulk_df = bulk_df.rename(columns = {'rearrangement':item_names[3]})
    else:
        selections = ['subject','productive_frequency', 'templates','epitope',item_names[0],item_names[1],item_names[2],'valid_cdr3','rearrangement'] + additional_cols
        bulk_df = bulk_df[selections].copy()
        bulk_df = bulk_df.rename(columns = {'rearrangement':item_names[3]})

    # Logging
    if return_valid_cdr3_only:
        bulk_df = bulk_df[bulk_df['valid_cdr3']]
        if log: logging.info(f"VALID CDR3 ({valid }) / ({bulk_df.shape[0]}) CLONES")
        if log: logging.info(f"OMITTING INVALID CDR3s FROM FINAL RESULTS")
        if log: logging.info(bulk_df[bulk_df.valid_cdr3 == False][['subject', "cdr3_b_aa"]])
        if log: logging.info(f"Final Reults has Column Names {bulk_df.columns}")
        if log: logging.info(f"Invalid ADAPTIVE V-gene names {invalid_v_names} not included\n")
        if log: logging.info(f"Invalid ADAPTIVE J-gene names {invalid_j_names} not included\n")
    
    # Assign count the productive_frequency based on the < use_as_count > argument
    bulk_df['count'] = bulk_df[count].copy()

    return bulk_df


def _valid_cdr3(cdr3):
    """ Return True iff all amino acids are part of standard amino acid list"""
    if not isinstance(cdr3, str):
        return False
    else:
        amino_acids = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']
        valid = np.all([aa in amino_acids for aa in cdr3])
        return valid


def ispublic(gr, var = "subject", n = 1):
    """
    Return True if a cluster public, defined as comprised of members from multiple individuals 
    or cell subsets (e.g., CD4/CD8) 
    
    Parameters
    ----------
    gr : group
        within pandas Data.Frame.groupby
    var : str
        variable name of class that group most transcend to be considered public
    m : int
        number of unique values of selected variable to be considered public
    
    Returns
    -------
    r : bool
        True if a cluster public
    
    """
    r = len(gr[var].value_counts()) > n
    if r:
        return 'public',len(gr[var].value_counts())
    else:
        return 'private',len(gr[var].value_counts())


def simple_cluster_index(   
                            pw_distances = None,
                            method = 'ward',
                            criterion = "distance",
                            t = 75):
    """
    Get 'cluster_index' 

    Parameters
    ----------

    t : int
        scipy.cluster.hierarchy.fcluster param t
    criterion : str 
        scipy.cluster.hierarchy.fcluster param criterion 
    method : str 
        scipy.cluster.linkage param method 
    t : int
        scipy.cluster.hierarcy.fcluster param t

    Notes
    -----
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.fcluster.html
    """

    pw_distances
    
    compressed_dmat = scipy.spatial.distance.squareform(pw_distances, force = "vector")
    Z = linkage(compressed_dmat, method = method)
    cluster_index = fcluster(Z, t = t, criterion = criterion)
    return cluster_index


def cluster_index_to_df(cluster_index):
    """
    Parameters
    ----------
    cluster_index : np.ndarray 

    Returns
    -------
    cluster_df : pd.DataFrame

    Notes
    -----
    
    cluster_df format:
    
    cluster_id                                          neighbors  K_neighbors
        4  [16, 25, 26, 29, 32, 50, 61, 68, 69, 94, 103, ...           24
        92  [35, 38, 41, 105, 131, 146, 181, 186, 189, 206...           18

    """
    dl = dict()
    for k,v in enumerate(cluster_index):
        dl.setdefault(v, [])
        dl[v].append(k)

    cluster_df = pd.DataFrame({'neighbors' : pd.Series(dl)}).sort_index().reset_index().rename(columns = {'index':'cluster_id'})
    cluster_df['K_neighbors'] = cluster_df.neighbors.str.len()
    cluster_df = cluster_df.sort_values(by = 'K_neighbors', ascending = False)
    
    return cluster_df


def get_centroid_seq(df, cdr3_name = 'cdr3_b_aa'):
    """
    Given a list of sequences, returns the sequence with the minimum 
    sum of distances to all other seqs in the list.

    Parameters
    ----------
    seqs : list
        list of strings (amino acid rep)
    metric : func
        defaults to pw.metrics.nw_hamming_metric

    Returns
    -------
    centroid_seq : str

    Example 
    -------
    >>> seqs = ['CASSEILAALGTQYF', 'CASSWTSRETQYF', 'CASSLAQETQYF', 'CASSLAPGDVSQYF', 'CASSWDQETQYF', 'CASSLWWDSGANVLTF', 'CASSLARTLSSGANVLTF', 'CASIPGTLFTFSGANVLTF', 'CASSFASSGANVLTF', 'CASSYRLLSGANVLTF']	
    >>> get_centroid_seq(seqs)
    'CASSFASSGANVLTF'

    Notes 
    -----
    In case of multiple occurrences of the minimum values, the indices 
    corresponding to the first occurrence are returned.
    """
    #import pwseqdist as pw
    #from scipy.spatial.distance import squareform
    seqs = df[cdr3_name]
    if len(seqs) < 3:
        return df.head(1)[cdr3_name], None, None, None

    metrics = {
            "cdr3_b_aa" : pw.metrics.nb_vector_tcrdist,
            "pmhc_b_aa" : pw.metrics.nb_vector_tcrdist,
            "cdr2_b_aa" : pw.metrics.nb_vector_tcrdist,
            "cdr1_b_aa" : pw.metrics.nb_vector_tcrdist}

    # Define weights
    weights = { "cdr3_b_aa" : 3,
                "pmhc_b_aa" : 1,
                "cdr2_b_aa" : 1,
                "cdr1_b_aa" : 1}

    kargs = {
                "cdr3_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}
    
    dmat = _pws(df = df,
        metrics = metrics, 
        weights = weights, 
        store = False,
        uniquify=False,
        kargs = kargs)
        
    dmat = dmat['tcrdist']

    dmat = dmat.astype(int)
    iloc_idx= dmat.sum(axis = 0).argmin()
    centroid_seq  = df[cdr3_name].to_list()[iloc_idx]
    loc_idx = df.index.to_list()[iloc_idx]
    return centroid_seq, dmat, iloc_idx, loc_idx

def get_centroid_seq_alpha(df, cdr3_name = 'cdr3_a_aa'):
    """
    Given a list of sequences, returns the sequence with the minimum 
    sum of distances to all other seqs in the list.

    Parameters
    ----------
    seqs : list
        list of strings (amino acid rep)
    metric : func
        defaults to pw.metrics.nw_hamming_metric

    Returns
    -------
    centroid_seq : str

    Example 
    -------
    >>> seqs = ['CASSEILAALGTQYF', 'CASSWTSRETQYF', 'CASSLAQETQYF', 'CASSLAPGDVSQYF', 'CASSWDQETQYF', 'CASSLWWDSGANVLTF', 'CASSLARTLSSGANVLTF', 'CASIPGTLFTFSGANVLTF', 'CASSFASSGANVLTF', 'CASSYRLLSGANVLTF']	
    >>> get_centroid_seq(seqs)
    'CASSFASSGANVLTF'

    Notes 
    -----
    In case of multiple occurrences of the minimum values, the indices 
    corresponding to the first occurrence are returned.
    """
    #import pwseqdist as pw
    #from scipy.spatial.distance import squareform
    seqs = df[cdr3_name]
    if len(seqs) < 3:
        return df.head(1)[cdr3_name], None, None, None

    metrics = {
            "cdr3_a_aa" : pw.metrics.nb_vector_tcrdist,
            "pmhc_a_aa" : pw.metrics.nb_vector_tcrdist,
            "cdr2_a_aa" : pw.metrics.nb_vector_tcrdist,
            "cdr1_a_aa" : pw.metrics.nb_vector_tcrdist}

    # Define weights
    weights = { "cdr3_a_aa" : 3,
                "pmhc_a_aa" : 1,
                "cdr2_a_aa" : 1,
                "cdr1_a_aa" : 1}

    kargs = {
                "cdr3_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}
    
    dmat = _pws(df = df,
        metrics = metrics, 
        weights = weights, 
        store = False,
        uniquify=False,
        kargs = kargs)
        
    dmat = dmat['tcrdist']

    dmat = dmat.astype(int)
    iloc_idx= dmat.sum(axis = 0).argmin()
    centroid_seq  = df[cdr3_name].to_list()[iloc_idx]
    loc_idx = df.index.to_list()[iloc_idx]
    return centroid_seq, dmat, iloc_idx, loc_idx



def bulk_adaptive_dataset_to_tcrdist3_clone_df( bulk_filename = None,
                                                minimum_file = None,
                                                maximum_file = None,
                                                organism = 'human',
                                                chains = ['beta'],
                                                epitope = None):
    """
    bulk_filename : str 
        input adaptive file e.g., "KHBR20-00150_TCRB.tsv"
    minimum_file : str
        minimum (CDR3,V,J,freq, templates,  only) output path e.g.,"KHBR20-00150_TCRB.tsv.tcrdist3.v_min.csv",
    maximum_file : str
        maximum (All CDRs, V,J, freq, subject, and all field in clone_df) output path "KHBR20-00150_TCRB.tsv.tcrdist3.v_max.csv",
    organism = 'human',
    chains = ['beta'],
    epitope = None)
    Bulk adaptive (Beta, Human) data to a tcrdist3 clone DataFrame. 

    2020 adaptive data has column bio_identity, productive_frequency and template. 

    * We will use productive frequency for counts.
    * We will only consider CDR3 with all valid amino acids
    """

    log = True
    logging.basicConfig(filename='prepbulk.log',level=logging.DEBUG, format='KMB:%(asctime)s\n\t %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')

    # Load only three columns 
    try:
        bulk_df = pd.read_csv(bulk_filename, sep= "\t", usecols = ['bio_identity', 'productive_frequency', 'templates'])
    except ValueError as e:
        raise Exception('Bulk Adpative TCR file was missing required columns') from e

    # Coerce strings to numeric
    bulk_df['productive_frequency'] = pd.to_numeric(bulk_df['productive_frequency'],errors='coerce')

    # Parse bio-identity
    ns= {0:"cdr3_b_aa", 1:"v_gene", 2:"j_gene"}
    cdr_v_j = bulk_df['bio_identity'].str.split("+", expand = True).\
        rename(columns = lambda x: ns[x])
    bulk_df[["cdr3_b_aa","v_gene","j_gene"]] = cdr_v_j

    # Convert Names from Adapative to IMGT
    bulk_df['v_b_gene'] = bulk_df['v_gene'].apply(lambda x : adaptive_to_imgt['human'].get(x))
    bulk_df['j_b_gene'] = bulk_df['j_gene'].apply(lambda x : adaptive_to_imgt['human'].get(x))

     # Validate CDR sequences
    bulk_df['valid_cdr3'] = bulk_df['cdr3_b_aa'].apply(lambda cdr3: _valid_cdr3(cdr3)) 
    # Count number of valid seqs
    valid = np.sum(bulk_df['valid_cdr3'])
    
    # Assign subject the filename
    bulk_df['subject'] = bulk_filename
    # Assign a blank epitope
    if epitope is None:
        bulk_df['epitope'] = 'X' 
    else: 
        bulk_df['epitope'] = epitope

    # Report % Valid
    print(f"VALID CDR3 ({valid }) / ({bulk_df.shape[0]}) CLONES")
    if log: logging.info(f"VALID CDR3 ({valid }) / ({bulk_df.shape[0]}) CLONES")
    print(f"OMITTING INVALID CDR3s FROM FINAL RESULTS")
    if log: logging.info(f"OMITTING INVALID CDR3s FROM FINAL RESULTS")
    # Log what was dropped
    print(bulk_df[bulk_df.valid_cdr3 == False][['subject', "cdr3_b_aa"]])
    if log: logging.info(bulk_df[bulk_df.valid_cdr3 == False][['subject', "cdr3_b_aa"]])
    print(f"Final Reults has Column Names {bulk_df.columns}")
    if log: logging.info(f"Final Reults has Column Names {bulk_df.columns}")

    bulk_df = bulk_df[['subject','productive_frequency', 'templates','epitope','cdr3_b_aa','v_b_gene','j_b_gene','valid_cdr3']].copy()
    bulk_df = bulk_df[bulk_df['valid_cdr3']]
    
    # Asign Count the productive_frequency
    bulk_df['count'] = bulk_df['productive_frequency'].copy()

    tr = TCRrep(cell_df = bulk_df, 
                organism = 'human', 
                chains = ['beta'], 
                compute_distances = False,
                infer_index_cols = True,
                deduplicate=False,
                cpus = 1,
                db_file = 'alphabeta_gammadelta_db.tsv')

    tr.clone_df[['cdr3_b_aa','v_b_gene','j_b_gene','productive_frequency','templates']].to_csv(minimum_file, sep = "\t", index = False)
    tr.clone_df.to_csv(maximum_file, sep = "\t", index = False)
    
    return tr.clone_df.copy()

def default_dist_clust_centroids(infile, cpus = 1, cdr3_b_aa_weight = 5, max_dist = 200):
    
    from tcrdist.repertoire import TCRrep
    
    print(infile)
    
    df = pd.read_csv(infile)
    df['count'] = 1

    tr = TCRrep(cell_df = df, 
                organism = 'human', 
                chains = ['beta'], 
                compute_distances = False,
                infer_index_cols = True,
                deduplicate=False,
                cpus = cpus,
                store_all_cdr=False,
                db_file = 'alphabeta_gammadelta_db.tsv')
    
    # Overweight CDR3B
    tr.weights_b['cdr3_b_aa'] = cdr3_b_aa_weight 

    icols = ['cell_type','subject','v_b_gene','j_b_gene','cdr3_b_aa','cdr3_b_nucseq','cdr1_b_aa','cdr2_b_aa','pmhc_b_aa']
    # Manually assign index cols    
    tr.index_cols = icols
    # Deduplicate
    tr.deduplicate()
    # Compute Distances
    tr.compute_distances()
    
    
    # Cluster based on the max_dist
    ci = simple_cluster_index(tr.pw_beta, t = max_dist)

    # Get a DataFrame for clusters (It is returned largest to smallest)
    ci_df = cluster_index_to_df(cluster_index = ci)  

    # Determine degree of public sharing in each cluster
    publicities = list()
    for i,r in ci_df.iterrows():
        clone_cluster_df = tr.clone_df.iloc[r['neighbors'],]
        publicity,n_subjects = ispublic(clone_cluster_df)
        publicities.append((publicity,n_subjects))
    # Store public status in a DataFrame
    public_df = pd.DataFrame(publicities).rename(columns = {0:'public',1:'n_subjects'})

    # Iterate through the ci_df DataFrame
    counter = 0
    centroids = list()
    cluster_df_list = list()
    for i,r in ci_df.iterrows():
        counter = counter + 1
        #print(r['neighbors'])
        clone_cluster_df = tr.clone_df.iloc[r['neighbors'],]
        cluster_df_list.append(clone_cluster_df )
        publicity = ispublic(clone_cluster_df)
        try:
            centroid, dmatrix, iloc_ind, loc_ind, = get_centroid_seq(df = tr.clone_df.iloc[r['neighbors'],])
            #print(centroid)
            #print(dmatrix)
            #print(np.max(dmatrix))
            #print(clone_cluster_df.v_b_gene.value_counts())
            #print(clone_cluster_df.j_b_gene.value_counts())
            #print(clone_cluster_df.subject.value_counts())
            #print(clone_cluster_df)
            #print("----")
            #print(clone_cluster_df.iloc[iloc_ind,])
            #print(centroid)
            assert clone_cluster_df.iloc[iloc_ind,]['cdr3_b_aa'] == centroid
            #print(loc_ind)
            #print(tr.clone_df.iloc[loc_ind,])
            assert tr.clone_df.iloc[loc_ind,]['cdr3_b_aa'] == centroid
            # store the centroid
            centroids.append( clone_cluster_df.iloc[iloc_ind,].reset_index(drop=True).copy())
        except Exception as e:
            print(e)
            print("fail")
            centroids.append( clone_cluster_df.iloc[0,].reset_index(drop=True).copy())

    # Save the Information in a centroids_df

    # renames --- {0: 'cell_type', 1: 'subject', 2: 'v_b_gene', 3: 'j_b_gene', 4: 'cdr3_b_aa', 5: 'cdr3_b_nucs, 6: 'cdr1_b_aa', 7: 'cdr2_b_aa', 8: 'pmhc_b_aa', 9: 'count', 10: 'clone_id'}
    renames = {i:v for i,v in enumerate(tr.clone_df.columns.to_list())}
    centroids_df = pd.DataFrame(centroids).rename(columns = renames)
    centroids_df['neighbors']     = ci_df['neighbors'].to_list()
    centroids_df['K_neighbors']   = ci_df['K_neighbors'].to_list()
    centroids_df['cluster_id']    = ci_df['cluster_id'].to_list()
    centroids_df['public']        = public_df['public'].to_list()
    centroids_df['n_subjects']    = public_df['n_subjects'].to_list()
    centroids_df['size_order']    = list(range(0,centroids_df.shape[0]))
    
    tr.centroids_df = centroids_df.copy()
    return tr


def get_basic_centroids(tr, max_dist = 200, look = False, cdr3_name = 'cdr3_b_aa'):

    # Cluster based on the max_dist
    ci = simple_cluster_index(tr.pw_beta, t = max_dist)

    # Get a DataFrame for clusters (It is returned largest to smallest)
    ci_df = cluster_index_to_df(cluster_index = ci)  

    # Determine degree of public sharing in each cluster
    publicities = list()
    for i,r in ci_df.iterrows():
        clone_cluster_df = tr.clone_df.iloc[r['neighbors'],]
        publicity,n_subjects = ispublic(clone_cluster_df)
        publicities.append((publicity,n_subjects))
    # Store public status in a DataFrame
    public_df = pd.DataFrame(publicities).rename(columns = {0:'public',1:'n_subjects'})

    # Iterate through the ci_df DataFrame
    counter = 0
    centroids = list()
    cluster_df_list = list()
    for i,r in ci_df.iterrows():
        counter = counter + 1
        
        clone_cluster_df = tr.clone_df.iloc[r['neighbors'],]
        cluster_df_list.append(clone_cluster_df )
        publicity = ispublic(clone_cluster_df)
        try:
            centroid, dmatrix, iloc_ind, loc_ind, = get_centroid_seq(df = tr.clone_df.iloc[r['neighbors'],])
            assert clone_cluster_df.iloc[iloc_ind,][ cdr3_name] == centroid
            assert tr.clone_df.iloc[loc_ind,][ cdr3_name] == centroid
            centroids.append( clone_cluster_df.iloc[iloc_ind,].reset_index(drop=True).copy())
            if look: 
                #print(dmatrix)
                print(r['neighbors'])
                print(f"MAX DISTANCE IN CLUSTER: {np.max(dmatrix)}" )
                print("----")
        except Exception as e:
            print("Using first sequence when not enough available to find centroid.\n")
            #print('FAIL')
            centroids.append( clone_cluster_df.iloc[0,].reset_index(drop=True).copy())


    # Save the Information in a centroids_df

    # renames --- {0: 'cell_type', 1: 'subject', 2: 'v_b_gene', 3: 'j_b_gene', 4: 'cdr3_b_aa', 5: 'cdr3_b_nucs, 6: 'cdr1_b_aa', 7: 'cdr2_b_aa', 8: 'pmhc_b_aa', 9: 'count', 10: 'clone_id'}
    renames = {i:v for i,v in enumerate(tr.clone_df.columns.to_list())}
    centroids_df = pd.DataFrame(centroids).rename(columns = renames)
    centroids_df['neighbors']     = ci_df['neighbors'].to_list()
    centroids_df['K_neighbors']   = ci_df['K_neighbors'].to_list()
    centroids_df['cluster_id']    = ci_df['cluster_id'].to_list()
    centroids_df['public']        = public_df['public'].to_list()
    centroids_df['n_subjects']    = public_df['n_subjects'].to_list()
    centroids_df['size_order']    = list(range(0,centroids_df.shape[0]))
    
    tr.centroids_df = centroids_df.copy()
    return tr



import numpy as np

"""
This module contains some basic code from the original tcrdist repo that 
supports .get_translation function call in repertoire_db

    self.protseq = translation.get_translation( self.nucseq, self.frame )[0]
"""

"""
from amino_acids.py module
"""
amino_acids = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', \
               'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']


longer_names={'ALA': 'A', 'ARG': 'R', 'ASN': 'N', 'ASP': 'D',
              'CYS': 'C', 'GLU': 'E', 'GLN': 'Q', 'GLY': 'G',
              'HIS': 'H', 'ILE': 'I', 'LEU': 'L', 'LYS': 'K',
              'MET': 'M', 'PHE': 'F', 'PRO': 'P', 'SER': 'S',
              'THR': 'T', 'TRP': 'W', 'TYR': 'Y', 'VAL': 'V'}

short_to_long = {}

for rsd in list(longer_names.keys()):short_to_long[longer_names[rsd]] = rsd

HP = {'I': 0.73, 'F': 0.61, 'V': 0.54, 'L': 0.53, 'W': 0.37,
      'M': 0.26, 'A': 0.25, 'G': 0.16, 'C': 0.04, 'Y': 0.02,
      'P': -0.07, 'T': -0.18, 'S': -0.26, 'H': -0.40, 'E': -0.62,
      'N': -0.64, 'Q': -0.69, 'D': -0.72, 'K': -1.10, 'R': -1.76}

HP['X'] = sum(HP.values())/20.

GES = {'F': -3.7, 'M': -3.4, 'I': -3.1, 'L': -2.8, 'V': -2.6,
       'C': -2.0, 'W': -1.9, 'A': -1.6, 'T': -1.2, 'G': -1.0,
       'S': -0.6, 'P': 0.2,  'Y': 0.7,  'H': 3.0,  'Q': 4.1,
       'N': 4.8,  'E': 8.2,  'K': 8.8,  'D': 9.2,  'R': 12.3}

GES['X'] = sum(GES.values())/20.

## KD values (Kyte-Doolittle) taken from http://web.expasy.org/protscale/pscale/Hphob.Doolittle.html

KD = {'A': 1.8, 'C': 2.5, 'E': -3.5, 'D': -3.5, 'G': -0.4, 'F': 2.8, 'I': 4.5, 'H': -3.2, 'K': -3.9, 'M': 1.9, 'L': 3.8, 'N': -3.5, 'Q': -3.5, 'P': -1.6, 'S': -0.8, 'R': -4.5, 'T': -0.7, 'W': -0.9, 'V': 4.2, 'Y': -1.3}
assert len(KD) == 20

aa_charge =  {}
for a in amino_acids: aa_charge[a] = 0.0
aa_charge['K'] = 1.0
aa_charge['R'] = 1.0
aa_charge['D'] = -1.0
aa_charge['E'] = -1.0
aa_charge['X'] = 0.0

groups = dict( zip( amino_acids, amino_acids ) )

groups['k'] = '[KR]'
groups['d'] = '[DE]'
groups['n'] = '[NQ]'
groups['s'] = '[ST]'
groups['f'] = '[FYWH]'
groups['a'] = '[AGSP]'
groups['v'] = '[VILM]'

"""
from genetic_code.py module
"""

## http://python.genedrift.org/2007/04/19/genetic-code-part-i/
gencode = {
    'ATA': 'I',    #Isoleucine
    'ATC': 'I',    #Isoleucine
    'ATT': 'I',    # Isoleucine
    'ATG': 'M',    # Methionine
    'ACA': 'T',    # Threonine
    'ACC': 'T',    # Threonine
    'ACG': 'T',    # Threonine
    'ACT': 'T',    # Threonine
    'AAC': 'N',    # Asparagine
    'AAT': 'N',    # Asparagine
    'AAA': 'K',    # Lysine
    'AAG': 'K',    # Lysine
    'AGC': 'S',    # Serine
    'AGT': 'S',    # Serine
    'AGA': 'R',    # Arginine
    'AGG': 'R',    # Arginine
    'CTA': 'L',    # Leucine
    'CTC': 'L',    # Leucine
    'CTG': 'L',    # Leucine
    'CTT': 'L',    # Leucine
    'CCA': 'P',    # Proline
    'CCC': 'P',    # Proline
    'CCG': 'P',    # Proline
    'CCT': 'P',    # Proline
    'CAC': 'H',    # Histidine
    'CAT': 'H',    # Histidine
    'CAA': 'Q',    # Glutamine
    'CAG': 'Q',    # Glutamine
    'CGA': 'R',    # Arginine
    'CGC': 'R',    # Arginine
    'CGG': 'R',    # Arginine
    'CGT': 'R',    # Arginine
    'GTA': 'V',    # Valine
    'GTC': 'V',    # Valine
    'GTG': 'V',    # Valine
    'GTT': 'V',    # Valine
    'GCA': 'A',    # Alanine
    'GCC': 'A',    # Alanine
    'GCG': 'A',    # Alanine
    'GCT': 'A',    # Alanine
    'GAC': 'D',    # Aspartic Acid
    'GAT': 'D',    # Aspartic Acid
    'GAA': 'E',    # Glutamic Acid
    'GAG': 'E',    # Glutamic Acid
    'GGA': 'G',    # Glycine
    'GGC': 'G',    # Glycine
    'GGG': 'G',    # Glycine
    'GGT': 'G',    # Glycine
    'TCA': 'S',    # Serine
    'TCC': 'S',    # Serine
    'TCG': 'S',    # Serine
    'TCT': 'S',    # Serine
    'TTC': 'F',    # Phenylalanine
    'TTT': 'F',    # Phenylalanine
    'TTA': 'L',    # Leucine
    'TTG': 'L',    # Leucine
    'TAC': 'Y',    # Tyrosine
    'TAT': 'Y',    # Tyrosine
    'TAA': '_',    # Stop
    'TAG': '_',    # Stop
    'TGC': 'C',    # Cysteine
    'TGT': 'C',    # Cysteine
    'TGA': '_',    # Stop
    'TGG': 'W',    # Tryptophan
    }


## http://www.cmbi.kun.nl/pythoncourse/spy/index.spy?site=python&action=Grand%20Finale&flag=chap

standard = { 'TTT': 'F', 'TCT': 'S', 'TAT': 'Y', 'TGT': 'C',
             'TTC': 'F', 'TCC': 'S', 'TAC': 'Y', 'TGC': 'C',
             'TTA': 'L', 'TCA': 'S', 'TAA': '*', 'TGA': 'W',
             'TTG': 'L', 'TCG': 'S', 'TAG': '*', 'TGG': 'W',
             'CTT': 'L', 'CCT': 'P', 'CAT': 'H', 'CGT': 'R',
             'CTC': 'L', 'CCC': 'P', 'CAC': 'H', 'CGC': 'R',
             'CTA': 'L', 'CCA': 'P', 'CAA': 'Q', 'CGA': 'R',
             'CTG': 'L', 'CCG': 'P', 'CAG': 'Q', 'CGG': 'R',
             'ATT': 'I', 'ACT': 'T', 'AAT': 'N', 'AGT': 'S',
             'ATC': 'I', 'ACC': 'T', 'AAC': 'N', 'AGC': 'S',
             'ATA': 'M', 'ACA': 'T', 'AAA': 'K', 'AGA': 'R',
             'ATG': 'M', 'ACG': 'T', 'AAG': 'K', 'AGG': 'R',
             'GTT': 'V', 'GCT': 'A', 'GAT': 'D', 'GGT': 'G',
             'GTC': 'V', 'GCC': 'A', 'GAC': 'D', 'GGC': 'G',
             'GTA': 'V', 'GCA': 'A', 'GAA': 'E', 'GGA': 'G',
             'GTG': 'V', 'GCG': 'A', 'GAG': 'E', 'GGG': 'G'}

## http://www.pasteur.fr/formation/infobio/python/ch15.html

code = {'ttt': 'F', 'tct': 'S', 'tat': 'Y', 'tgt': 'C',
        'ttc': 'F', 'tcc': 'S', 'tac': 'Y', 'tgc': 'C',
        'tta': 'L', 'tca': 'S', 'taa': '*', 'tga': '*',
        'ttg': 'L', 'tcg': 'S', 'tag': '*', 'tgg': 'W',
        'ctt': 'L', 'cct': 'P', 'cat': 'H', 'cgt': 'R',
        'ctc': 'L', 'ccc': 'P', 'cac': 'H', 'cgc': 'R',
        'cta': 'L', 'cca': 'P', 'caa': 'Q', 'cga': 'R',
        'ctg': 'L', 'ccg': 'P', 'cag': 'Q', 'cgg': 'R',
        'att': 'I', 'act': 'T', 'aat': 'N', 'agt': 'S',
        'atc': 'I', 'acc': 'T', 'aac': 'N', 'agc': 'S',
        'ata': 'I', 'aca': 'T', 'aaa': 'K', 'aga': 'R',
        'atg': 'M', 'acg': 'T', 'aag': 'K', 'agg': 'R',
        'gtt': 'V', 'gct': 'A', 'gat': 'D', 'ggt': 'G',
        'gtc': 'V', 'gcc': 'A', 'gac': 'D', 'ggc': 'G',
        'gta': 'V', 'gca': 'A', 'gaa': 'E', 'gga': 'G',
        'gtg': 'V', 'gcg': 'A', 'gag': 'E', 'ggg': 'G'
        }

genetic_code = {}
reverse_genetic_code = {}

for codon in gencode:
    lowcodon = codon.lower()
    assert code[ lowcodon ] == gencode[ codon ] or ( gencode[ codon ] == '_' and code[ lowcodon ] == '*' )
    aa = code[ lowcodon ]
    genetic_code[ lowcodon ] = aa
    if aa not in reverse_genetic_code: reverse_genetic_code[aa] = []
    reverse_genetic_code[aa].append( lowcodon )


"""
From translation.py module
"""
base_partner = {'a': 't', 'c': 'g', 'g': 'c', 't': 'a', 'n': 'n',
                'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N',
                'R': 'Y', 'Y': 'R',
                'S': 'S', 'W': 'W',
                'K': 'M', 'M': 'K',
                '.': '.' }

nucleotide_classes_lower_case = { 'a':'a',
                                  'c':'c',
                                  'g':'g',
                                  't':'t',
                                  'w':'at',
                                  's':'cg',
                                  'k':'gt',
                                  'm':'ac',
                                  'y':'ct',
                                  'r':'ag',
                                  'n':'acgt' }

def reverse_complement( seq ):
    newseq = ''
    L = len(seq)
    for pos in range( L-1, -1, -1 ):
        newseq += base_partner[ seq[ pos ] ]
    assert len( newseq ) == L
    return newseq

def modify_genetic_code(genetic_code):
    bases_plus = list(nucleotide_classes_lower_case.keys())

    for a in bases_plus:
        for b in bases_plus:
            for c in bases_plus:
                codon = a+b+c
                if codon in genetic_code: continue

                aas = []
                for a1 in nucleotide_classes_lower_case[a]:
                    for b1 in nucleotide_classes_lower_case[b]:
                        for c1 in nucleotide_classes_lower_case[c]:
                            aas.append( genetic_code[ a1+b1+c1 ] )
                if min(aas) == max(aas):
                    genetic_code[codon] = aas[0]
                else:
                    genetic_code[codon] = 'X'
    return genetic_code

def get_translation(seq, frame):
    assert frame in [-1, -2, -3, 1, 2, 3, '+1', '+2', '+3', '1', '2', '3', '-1', '-2', '-3']
    frame = int(frame)
    if frame < 0:
        seq = reverse_complement( seq )
    offset = np.abs(frame) - 1
    assert offset in range(3)
    seq = seq[offset:].lower()
    naa = len(seq)//3
    protseq = ''
    codons = []
    for i in range(int(naa)):
        codon = seq[3*i:3*i+3]
        codons.append( codon )
        if '#' in codon:
            protseq += '#'
        else:
            protseq += genetic_code.get( codon, 'X' )
    return protseq, codons

genetic_code = modify_genetic_code(genetic_code)




import pytest



def test_example_8():
    """
    You want a 'tcrdistance' but you don't want to bother with the tcrdist3 framework. 
    
    Note that the columns names are completely arbitrary under this 
    framework, so one can directly compute a tcrdist on a 
    AIRR, MIXCR, VDJTools, or other formated file without any
    reformatting.
    """ 
    import multiprocessing
    import pandas as pd
    import pwseqdist as pw
    from tcrdist.rep_funcs import _pws, _pw  
    
    df_airr = pd.read_csv("dash_beta_airr.csv")

    # Choose the metrics you want to apply to each CDR
    metrics = { 'cdr3_aa' : pw.metrics.nb_vector_tcrdist,
                'cdr2_aa' : pw.metrics.nb_vector_tcrdist,
                'cdr1_aa' : pw.metrics.nb_vector_tcrdist}
    
    # Choose the weights that are right for you.
    weights = { 'cdr3_aa' : 3,
                'cdr2_aa' : 1,
                'cdr1_aa' : 1}

    # Provide arguments for the distance metrics 
    kargs = {   'cdr3_aa' : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                'cdr2_aa' : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                'cdr1_aa' : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}
                
    # Here are your distance matrices
    from tcrdist.rep_funcs import _pws
    
    dmats = _pws(df = df_airr,
             metrics = metrics, 
             weights= weights, 
             kargs=kargs, 
             cpu = 1, 
             store = True)

    dmats['tcrdist']



"""
test_breadth.py 

The breadth and depth of a disease-specific T-cell response.

This module concerns the estimation of clonal breadth, whether it be at the 
pathogen, protein, or epitope level. Once meta-clonotype have been defined they 
can be used to search for biochemically similar TCRs in bulk repertoires 
that are likely to share antigen recognition. It is possible that a
single TCR clonotype may be conformant with multiple TCR meta-clonotypes, 
so an accurate estimate of clonal breadth must avoid double counting such 
clonotypes.  

To estimate clonal breadth of antigen-associated TCRs within 
a bulk repertoires with (N) productive clonotypes and (M) total 
productive templates, we use a set of (X) previously defined
antigen-associated meta-clonotypes (defined as a (i) Centroid TRV,CDR3, 
(ii) TCR-specific RADIUS, (iii) MOTIF. 

1. Compute the TCRdist between each centroid TCRij for i {1...i...X} 
and all bulk clones {1...j..N} using rectangular search with the 
tcrdist.repertoires.TCRrep.compute_sparse_rect_distances(), producing 
a sparse distance matrix. 

2. Next perform a long-form tabulation that records the network formed between 
all meta-clonotype centroids and bulk sequences within the specified radius. 
This is performed with the function tcrdist.breadth.long_form_tabulation().

The network is represented as a Pandas DataFrame. Where centroid sequences
are recorded as "cdr3_b_aa", "v_b_gene", "j_b_gene" and the conformant sequence
in the bulk repertoire is "cdr3_b_aa_hit", 'v_b_gene_hit', 'j_b_gene_hit'. 
Crucially there is a column "MOTIF" which indicates whether the CDR3 of 
the hit sequence is conformant with the regular expression in the column 
"regex". 

3. The long-form Pandas DataFrame can then be used as input to the function 
tcrdist.breadth.estimate_breadth_and_depth(). The unit of analysis -- that is, 
whether breadth refers to pathogen, protein, or epitope specific breadth --
can be specified in with the argument <breadth_cols>. Crucially, when 
running tcrdist.breadth.long_form_tabulation() the argument 
<search_cols> must include a column indicating the association between 
a metac-clonotype and a particular 'protein' or 'epitope'  
e.g., ['tag', 'protein', 'epitope', cdr3_b_aa', 'v_b_gene', 'j_b_gene', 
'pgen','regex', 'radius']

Note that breadth and depth follow the definitions in Synder et al. 2020 
(https://www.ncbi.nlm.nih.gov/pmc/articles/PMC7418734/), 
see section: The breadth and depth of a disease-specific T-cell response. 

"""
import pytest

#Download many bulk repertoires or just 2
def test_breadth_estimation_workflow(testing_only = True):
    import numpy as np
    import os 
    import pandas as pd
    from tcrdist.setup_tests import download_and_extract_zip_file
    from tcrdist.repertoire import TCRrep
    from tcrdist.breadth import get_safe_chunk
    from tcrdist.breadth import long_form_tabulation
    from tcrdist.breadth import estimate_breadth_and_depth
    import multiprocessing
    import time
    import scipy.sparse
    # For testing use only a subset of meta-clonotypes
  
    ncpus = min(multiprocessing.cpu_count(), 6)
    # Download 2 bulk files for testing and demonstration purposes
    files = [
    '1588BW_20200417_PBMC_unsorted_cc1000000_ImmunRACE_050820_008_gDNA_TCRB.tsv.tcrdist3.tsv',
    '1349BW_unsorted_cc1000000_ImmunRACE_043020_003_gDNA_TCRB.tsv.tcrdist3.tsv']
    if not np.all([os.path.isfile(f) for f in files]):
        download_and_extract_zip_file(
            "ImmunoSeq_MIRA_matched_tcrdist3_ready_2_files.zip", 
            source = "dropbox", 
            dest = ".")
        assert np.all([os.path.isfile(f) for f in files])
    # Download a Meta-Clonotypes File, All Meta-Clonotypes from Pre-Print Manuscript 
    # (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC7781332/)
    if not os.path.isfile("bioRxiv_v2_metaclonotypes.tsv.zip"): 
        download_and_extract_zip_file('bioRxiv_v2_metaclonotypes.tsv.zip', source = "dropbox", dest = ".")
        assert os.path.isfile('bioRxiv_v2_metaclonotypes.tsv')
    # ███████╗███████╗ █████╗ ██████╗  ██████╗██╗  ██╗
    # ██╔════╝██╔════╝██╔══██╗██╔══██╗██╔════╝██║  ██║
    # ███████╗█████╗  ███████║██████╔╝██║     ███████║
    # ╚════██║██╔══╝  ██╔══██║██╔══██╗██║     ██╔══██║
    # ███████║███████╗██║  ██║██║  ██║╚██████╗██║  ██║
    # ╚══════╝╚══════╝╚═╝  ╚═╝╚═╝  ╚═╝ ╚═════╝╚═╝  ╚═╝
    meta_clonotypes_filename = 'bioRxiv_v2_metaclonotypes.tsv'
    meta_clonotypes          = pd.read_csv(meta_clonotypes_filename , sep = "\t")
    # Assign feature id as a tag for tracking by meta-clonotype definition
    meta_clonotypes['tag'] = meta_clonotypes['feature'].copy()
    cols = ['cdr3_b_aa','v_b_gene','j_b_gene',
            'pgen','radius','regex','protein','protein_coordinate', 'tag']
    df_search = meta_clonotypes[cols]
    df_search = df_search.assign(count= 1)

    if testing_only: 
        # take a subsample to accelerated unit testing
        df_search = df_search.sample(100, random_state=1)
    tr_search = TCRrep(cell_df = df_search,
                    organism = 'human',
                    chains = ['beta'],
                    db_file = 'alphabeta_gammadelta_db.tsv',
                    compute_distances = False)
    tr_search.cpus = ncpus
    # ██████╗ ██╗   ██╗██╗     ██╗  ██╗
    # ██╔══██╗██║   ██║██║     ██║ ██╔╝
    # ██████╔╝██║   ██║██║     █████╔╝ 
    # ██╔══██╗██║   ██║██║     ██╔═██╗ 
    # ██████╔╝╚██████╔╝███████╗██║  ██╗
    # ╚═════╝  ╚═════╝ ╚══════╝╚═╝  ╚═╝
    bulk_filename = files[0]
    df_bulk = pd.read_csv( bulk_filename, sep = "\t")
    valid_TRBV_CDR3_TRBJ = (df_bulk['valid_cdr3'] == True)&(df_bulk['v_b_gene'].notna())&(df_bulk['j_b_gene'].notna())
    df_bulk = df_bulk[valid_TRBV_CDR3_TRBJ].reset_index(drop = True)
    # Convert templates to counts
    if 'templates' in df_bulk.columns:
        df_bulk = df_bulk[['cdr3_b_aa','v_b_gene','j_b_gene','templates',
        'productive_frequency']].\
        rename(columns = {'templates':'count'})
    else:
        df_bulk = df_bulk[['cdr3_b_aa','v_b_gene','j_b_gene','count','productive_frequency']]
    tr_bulk   = TCRrep(cell_df = df_bulk,
                    organism = 'human',
                    chains = ['beta'],
                    db_file = 'alphabeta_gammadelta_db.tsv',
                    compute_distances = False)
    # Get dimensions
    search_clones = tr_search.clone_df.shape[0] 
    bulk_clones   = tr_bulk.clone_df.shape[0]
    # Get and ideal chunksize that will control memory usage
    # 10**7 will keep memory > 2GB per CPU. 
    ideal_chunk_size = get_safe_chunk(
        search_clones = tr_search.clone_df.shape[0], 
        bulk_clones = tr_bulk.clone_df.shape[0],
        target = 10**7)
    print(f"IDEAL CHUNK SIZE {ideal_chunk_size},{ncpus} CPUS")
    # ██████╗ ███████╗ ██████╗████████╗    ███████╗███████╗ █████╗ ██████╗  ██████╗██╗  ██╗
    # ██╔══██╗██╔════╝██╔════╝╚══██╔══╝    ██╔════╝██╔════╝██╔══██╗██╔══██╗██╔════╝██║  ██║
    # ██████╔╝█████╗  ██║        ██║       ███████╗█████╗  ███████║██████╔╝██║     ███████║
    # ██╔══██╗██╔══╝  ██║        ██║       ╚════██║██╔══╝  ██╔══██║██╔══██╗██║     ██╔══██║
    # ██║  ██║███████╗╚██████╗   ██║       ███████║███████╗██║  ██║██║  ██║╚██████╗██║  ██║
    # ╚═╝  ╚═╝╚══════╝ ╚═════╝   ╚═╝       ╚══════╝╚══════╝╚═╝  ╚═╝╚═╝  ╚═╝ ╚═════╝╚═╝  ╚═╝
    tic = time.perf_counter()
    tr_search.compute_sparse_rect_distances(
        df = tr_search.clone_df, 
        df2 = tr_bulk.clone_df, 
        chunk_size = ideal_chunk_size, 
        radius = 37)
    toc = time.perf_counter()
    print(f"SEARCHING {search_clones} META-CLONOTYPES IN {bulk_clones} BULK CLONES IN {toc - tic:0.2f} sec.")
    # ████████╗ █████╗ ██████╗ ██╗   ██╗██╗      █████╗ ████████╗███████╗
    # ╚══██╔══╝██╔══██╗██╔══██╗██║   ██║██║     ██╔══██╗╚══██╔══╝██╔════╝
    #    ██║   ███████║██████╔╝██║   ██║██║     ███████║   ██║   █████╗  
    #    ██║   ██╔══██║██╔══██╗██║   ██║██║     ██╔══██║   ██║   ██╔══╝  
    #    ██║   ██║  ██║██████╔╝╚██████╔╝███████╗██║  ██║   ██║   ███████╗
    #    ╚═╝   ╚═╝  ╚═╝╚═════╝  ╚═════╝ ╚══════╝╚═╝  ╚═╝   ╚═╝   ╚══════╝
    df_tab = long_form_tabulation(
        clone_df1   = tr_search.clone_df,
        clone_df2   = tr_bulk.clone_df,
        csrmat      = tr_search.rw_beta, 
        search_cols = ['tag', 'protein','protein_coordinate','cdr3_b_aa', 
        'v_b_gene', 'j_b_gene', 'pgen','regex', 'radius'])
    # ██████╗ ██████╗ ███████╗ █████╗ ██████╗ ████████╗██╗  ██╗
    # ██╔══██╗██╔══██╗██╔════╝██╔══██╗██╔══██╗╚══██╔══╝██║  ██║
    # ██████╔╝██████╔╝█████╗  ███████║██║  ██║   ██║   ███████║
    # ██╔══██╗██╔══██╗██╔══╝  ██╔══██║██║  ██║   ██║   ██╔══██║
    # ██████╔╝██║  ██║███████╗██║  ██║██████╔╝   ██║   ██║  ██║
    # ╚═════╝ ╚═╝  ╚═╝╚══════╝╚═╝  ╚═╝╚═════╝    ╚═╝   ╚═╝  ╚═╝
    df_breadth = estimate_breadth_and_depth(df=df_tab, 
        breadth_cols = ['protein'], 
        N =tr_bulk.clone_df.shape[0] , 
        M= tr_bulk.clone_df['count'].sum(), 
        motif = True,
        exact = False)
    df_breadth = df_breadth.assign(file = files[0])
    # ███╗   ███╗ █████╗ ███╗   ██╗██╗   ██╗    ███████╗██╗██╗     ███████╗███████╗
    # ████╗ ████║██╔══██╗████╗  ██║╚██╗ ██╔╝    ██╔════╝██║██║     ██╔════╝██╔════╝
    # ██╔████╔██║███████║██╔██╗ ██║ ╚████╔╝     █████╗  ██║██║     █████╗  ███████╗
    # ██║╚██╔╝██║██╔══██║██║╚██╗██║  ╚██╔╝      ██╔══╝  ██║██║     ██╔══╝  ╚════██║
    # ██║ ╚═╝ ██║██║  ██║██║ ╚████║   ██║       ██║     ██║███████╗███████╗███████║
    # ╚═╝     ╚═╝╚═╝  ╚═╝╚═╝  ╚═══╝   ╚═╝       ╚═╝     ╚═╝╚══════╝╚══════╝╚══════╝
    import numpy as np
    import os 
    import pandas as pd
    from tcrdist.setup_tests import download_and_extract_zip_file
    from tcrdist.repertoire import TCRrep
    from tcrdist.breadth import get_safe_chunk
    from tcrdist.breadth import long_form_tabulation
    from tcrdist.breadth import estimate_breadth_and_depth
    import multiprocessing
    import time
    import scipy.sparse
    ncpus = min(multiprocessing.cpu_count(), 6)
    destination = "breadth_estimates"
    if not os.path.isdir(destination):
        os.mkdir(destination)
    files = [
    '1588BW_20200417_PBMC_unsorted_cc1000000_ImmunRACE_050820_008_gDNA_TCRB.tsv.tcrdist3.tsv',
    '1349BW_unsorted_cc1000000_ImmunRACE_043020_003_gDNA_TCRB.tsv.tcrdist3.tsv']

    # ███████╗███████╗ █████╗ ██████╗  ██████╗██╗  ██╗
    # ██╔════╝██╔════╝██╔══██╗██╔══██╗██╔════╝██║  ██║
    # ███████╗█████╗  ███████║██████╔╝██║     ███████║
    # ╚════██║██╔══╝  ██╔══██║██╔══██╗██║     ██╔══██║
    # ███████║███████╗██║  ██║██║  ██║╚██████╗██║  ██║
    # ╚══════╝╚══════╝╚═╝  ╚═╝╚═╝  ╚═╝ ╚═════╝╚═╝  ╚═╝
    meta_clonotypes_filename = 'bioRxiv_v2_metaclonotypes.tsv'
    meta_clonotypes          = pd.read_csv(meta_clonotypes_filename , sep = "\t")
    # Assign feature id as a tag for tracking by meta-clonotype definition
    meta_clonotypes['tag'] = meta_clonotypes['feature'].copy()
    cols = ['cdr3_b_aa','v_b_gene','j_b_gene',
            'pgen','radius','regex','protein','protein_coordinate', 'tag']
    df_search = meta_clonotypes[cols]
    df_search = df_search.assign(count= 1)
    if testing_only: 
        # take a subsample to accelerated unit testing
        df_search = df_search.sample(100, random_state=1)
    tr_search = TCRrep(cell_df = df_search,
                    organism = 'human',
                    chains = ['beta'],
                    db_file = 'alphabeta_gammadelta_db.tsv',
                    compute_distances = False)
    tr_search.cpus = ncpus

    for file in files:
        bulk_filename = file
        df_bulk = pd.read_csv( bulk_filename, sep = "\t")
        valid_TRBV_CDR3_TRBJ = (df_bulk['valid_cdr3'] == True)&(df_bulk['v_b_gene'].notna())&(df_bulk['j_b_gene'].notna())
        df_bulk = df_bulk[valid_TRBV_CDR3_TRBJ].reset_index(drop = True)
        # Convert templates to counts
        if 'templates' in df_bulk.columns:
            df_bulk = df_bulk[['cdr3_b_aa','v_b_gene','j_b_gene','templates','productive_frequency']].rename(columns = {'templates':'count'})
        else:
            df_bulk = df_bulk[['cdr3_b_aa','v_b_gene','j_b_gene','count','productive_frequency']]
        tr_bulk   = TCRrep(cell_df = df_bulk,
                        organism = 'human',
                        chains = ['beta'],
                        db_file = 'alphabeta_gammadelta_db.tsv',
                        compute_distances = False)
        # Get dimensions
        search_clones = tr_search.clone_df.shape[0] 
        bulk_clones   = tr_bulk.clone_df.shape[0]
        # Get and ideal chunksize that will control memory usage
        # 10**7 will keep memory > 2GB per CPU. 
        ideal_chunk_size = get_safe_chunk(tr_search.clone_df.shape[0], tr_bulk.clone_df.shape[0], target = 10**7)
        print(f"IDEAL CHUNK SIZE {ideal_chunk_size},{ncpus} CPUS")
        # ██████╗ ███████╗ ██████╗████████╗    ███████╗███████╗ █████╗ ██████╗  ██████╗██╗  ██╗
        # ██╔══██╗██╔════╝██╔════╝╚══██╔══╝    ██╔════╝██╔════╝██╔══██╗██╔══██╗██╔════╝██║  ██║
        # ██████╔╝█████╗  ██║        ██║       ███████╗█████╗  ███████║██████╔╝██║     ███████║
        # ██╔══██╗██╔══╝  ██║        ██║       ╚════██║██╔══╝  ██╔══██║██╔══██╗██║     ██╔══██║
        # ██║  ██║███████╗╚██████╗   ██║       ███████║███████╗██║  ██║██║  ██║╚██████╗██║  ██║
        # ╚═╝  ╚═╝╚══════╝ ╚═════╝   ╚═╝       ╚══════╝╚══════╝╚═╝  ╚═╝╚═╝  ╚═╝ ╚═════╝╚═╝  ╚═╝
        tic = time.perf_counter()
        tr_search.compute_sparse_rect_distances(
            df = tr_search.clone_df, 
            df2 = tr_bulk.clone_df, 
            chunk_size = ideal_chunk_size, 
            radius = 37)
        toc = time.perf_counter()
        print(f"SEARCHING {search_clones} META-CLONOTYPES IN {bulk_clones} BULK CLONES IN {toc - tic:0.2f} sec.")
        # ████████╗ █████╗ ██████╗ ██╗   ██╗██╗      █████╗ ████████╗███████╗
        # ╚══██╔══╝██╔══██╗██╔══██╗██║   ██║██║     ██╔══██╗╚══██╔══╝██╔════╝
        #    ██║   ███████║██████╔╝██║   ██║██║     ███████║   ██║   █████╗  
        #    ██║   ██╔══██║██╔══██╗██║   ██║██║     ██╔══██║   ██║   ██╔══╝  
        #    ██║   ██║  ██║██████╔╝╚██████╔╝███████╗██║  ██║   ██║   ███████╗
        #    ╚═╝   ╚═╝  ╚═╝╚═════╝  ╚═════╝ ╚══════╝╚═╝  ╚═╝   ╚═╝   ╚══════╝
        df_tab = long_form_tabulation(
            clone_df1   = tr_search.clone_df,
            clone_df2   = tr_bulk.clone_df,
            csrmat      = tr_search.rw_beta, 
            search_cols = ['tag', 'protein','protein_coordinate','cdr3_b_aa', 
            'v_b_gene', 'j_b_gene', 'pgen','regex', 'radius'])
        
        # ██████╗ ██████╗ ███████╗ █████╗ ██████╗ ████████╗██╗  ██╗
        # ██╔══██╗██╔══██╗██╔════╝██╔══██╗██╔══██╗╚══██╔══╝██║  ██║
        # ██████╔╝██████╔╝█████╗  ███████║██║  ██║   ██║   ███████║
        # ██╔══██╗██╔══██╗██╔══╝  ██╔══██║██║  ██║   ██║   ██╔══██║
        # ██████╔╝██║  ██║███████╗██║  ██║██████╔╝   ██║   ██║  ██║
        # ╚═════╝ ╚═╝  ╚═╝╚══════╝╚═╝  ╚═╝╚═════╝    ╚═╝   ╚═╝  ╚═╝
        df_breadth = estimate_breadth_and_depth(df=df_tab, 
            breadth_cols = ['protein'], 
            N =tr_bulk.clone_df.shape[0] , 
            M= tr_bulk.clone_df['count'].sum(), 
            motif = True,
            exact = False)
        df_breadth = df_breadth.assign(file = files[0])
        # save rectangular sparse matrix, tabulation, and breadth esimates for future reference
        scipy.sparse.save_npz(os.path.join(destination, f"{file}.rw.npz"), tr_search.rw_beta)
        df_tab.to_csv(os.path.join(destination, f"{file}.tabulation.tsv"), 
            sep = '\t', index = True)
        df_breadth.to_csv(os.path.join(destination, f"{file}.breadth.tsv"),
            sep = '\t', index = True)



import pytest
import Levenshtein
def my_own_metric(s1,s2):   
    return Levenshtein.distance(s1,s2)
def test_example_6():
    """
    If you hate object oriented programming, just show me the functions. 
    No problem. 
    
    Maybe you only care about the CDR3 on the beta chain.

    def my_own_metric(s1,s2):   
        return Levenshtein.distance(s1,s2)
    """  
    import multiprocessing
    import pandas as pd
    from tcrdist.rep_funcs import _pws, _pw

    df = pd.read_csv("dash2.csv")

    # 
    dmat = _pw( metric = my_own_metric,
                seqs1 = df['cdr3_b_aa'].values,
                ncpus=2,
                uniqify=True,
                use_numba=False)



import pandas as pd
import pwseqdist as pw
import numpy as np
import warnings
from . import repertoire_db
from tcrdist.rep_funcs import _pws, compute_pws_sparse
from tcrdist.sample import _default_sampler
from tcrdist.background import get_stratified_gene_usage_frequency
from tcrdist.background import _synthesize_human_beta_vj_background, _synthesize_human_alpha_vj_background, _synthesize_mouse_beta_vj_background, _synthesize_mouse_alpha_vj_background
from zipdist.zip2 import Zipdist2
import sys


class TCRrep:
    """
    Flexible distance measures for comparing T cell receptors

    The TCRrep Class hold T cell repertoire data, infers CDRs from v-gene name, and 
    computes multi-CDR 'tcrdistance'.  

    Attributes
    ----------
    cell_df : pd.DataFrame or None
        Pandas DataFrame containing cell level information 
    clone_df : pd.DataFrame or None
        Pandas DataFrame containing clone level information. 
        This can be provided directly from a program like MIXCR or
        can be inferred by deduplicating a cell_df. 
    organism = str, 
        specifies relevant organism for analysis: 'human' or 'mouse'
    chains : list
        specifies relevant chains for single or paried TCR analysis
        ['alpha','beta'], ['alpha'], ['beta'], ['gamma','delta'],  ['gamma'] or ['delta']  
    db_file : str
        specifies reference file. The default is 'alphabeta_gammadelta_db.tsv' which
        is preinstalled  with the install python3.7/site-packages/tcrdist/db/
    archive_name : str
        Name for archive file. (only used if archive result is True)
    archive_result: bool
        If True save result to .tar.gz archive
    imgt_aligned : bool
        If True, by default, cdr1, cdr2,and pmhc are inferred aligned to fixed length with gaps.
        If False, cdr1, cdr2,and pmhc are returned as ungapped raw sequences.   
    infer_all_genes : bool
        If True, load all_gene reference from 'db_file`.
    infer_cdrs : bool
        If True, infer cdr1, cdr2,and pmhc from the v gene name
    infer_index_cols : bool
        If True, infer index_cols used to deduplicate cell_df.
        If False, index_cols can be specified directly after initialization.
    index_cols : list
        list of index columns used to deduplicate cell_df to clone_df
    deduplicate : bool
        If True, then clone_df will be assigned cell_df grouped_by
        index_cols.
        If False, and clone_df is None, then clone_df will be be 
        assigned a copy of cell_df.
    use_defaults : True
        If True, use default metrics, weights, and kargs
    store_all_cdr : True,
        If True, save all distance matrices for each CDR (e.g., pw_cdr3_b_aa).
        If False, only save pw_alpha and pw_beta
    compute_distances : True
        If True, automatically compute distances
    cpus : int, 
        Number of cpus to use. In general 1 cpu is sufficient from default Numba metric 
        with less than 10^7  pairwise comparisons. However, more cpus will 
        result in a speedup for metrics like pw.metrics.nw_hamming_metric for more than 10^6 
        pairwise comparisons.
    
    Example
    -------
    .. code-block:: python

        import pandas as pd
        from tcrdist.repertoire import TCRrep

        df = pd.read_csv("dash.csv")
        tr = TCRrep(
                cell_df = df, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv',
                clone_df          = None,
                imgt_aligned      = True,
                infer_all_genes   = True,
                infer_cdrs        = True,
                infer_index_cols  = True,
                deduplicate       = True,
                use_defaults      = True,
                store_all_cdr     = True,
                compute_distances = True,
                index_cols        = None,
                cpus              = 1, 
                db_file           = 'alphabeta_gammadelta_db.tsv')

    Notes
    -----
    The default initialization compute pairwise distance matrices. 

    See examples at https://tcrdist3.readthedocs.io/ for more information
    about flexibility of TCRrep.

    Computed pairwised distance matrices that can be accessed:
        TCRrep.pw_alpha, 
        TCRrep.pw_beta,  
        TCRrep.pw_cdr3_a_aa, 
        TCRrep.pw_cdr3_b_aa  
    """
    def __init__(self,
                 organism          = "mouse",
                 chains            = ['alpha', 'beta'],
                 db_file           = 'combo_xcr_2024-03-05.tsv',
                 archive_name      = 'tcrdist3.archive',
                 blank             = False,
                 cell_df           = None,
                 clone_df          = None,
                 imgt_aligned      = True,
                 infer_all_genes   = True,
                 infer_cdrs        = True,
                 infer_index_cols  = True,
                 deduplicate       = True,
                 use_defaults      = True,
                 store_all_cdr     = True,
                 compute_distances = True,
                 index_cols        = None,
                 cpus              = 1,
                 df2               = None, 
                 archive_result    = False):
        

        self.organism = organism
        self._validate_organism()   

        self.chains = chains
        self._validate_chains()
        
        self.archive_name = archive_name
        self.blank = blank 
        # blank is only used if reloading from .tar.gz
        if self.blank:
            self.cell_df           = None
            self.clone_df          = None
            self.imgt_aligned      = False
            self.infer_all_genes   = False
            self.store_all_cdr     = False
            self.index_cols        = None
            self.cpus              = 1,
            self.df2               = None 
            self.archive_result    = False
        else:
            self.index_cols = index_cols
            self.clone_df = clone_df 

            if cell_df is None:
                cell_df = pd.DataFrame()   
            self.cell_df = cell_df            
           
            self._validate_cell_df()
            self.clone_df = clone_df
            self.df2 = df2
            
            
            self.db_file = db_file
            self._validate_db_file()

            self.imgt_aligned = imgt_aligned
            self._validate_imgt_aligned()

            self.store_all_cdr = store_all_cdr
            self.cpus = cpus

            self.archive_result = archive_result
            self.infer_all_genes = infer_all_genes
            self.use_defaults = use_defaults

            if infer_all_genes:     
                self.all_genes = repertoire_db.RefGeneSet(db_file).all_genes
            
            if infer_cdrs:
                for chain in self.chains:
                    self.infer_cdrs_from_v_gene(chain = chain, imgt_aligned = self.imgt_aligned)
                        # Assume all provided columns are index columns, except 'count' 'cell_id', 'clone_id'
            
            if infer_index_cols:
                self.infer_index_cols()
            
            if deduplicate:
                self.infer_index_cols()
                self.deduplicate()
            else: 
                if self.clone_df is None:
                    self.clone_df = self.cell_df.copy()

            if use_defaults:
                self._initialize_chain_specific_attributes()
        
            if compute_distances:
                # This is a safety, measure so that a new user doesn't accidently try to compute a pairwise matrix that won't fit in memory
                if self.clone_df.shape[0] > 10000:
                    warnings.warn(f"\n\nWhen TCRrep.<clone_df> size {self.clone_df.shape[0]} > 10,000.\n"
                    "\tTCRrep.compute_distances() may be called explicitly by a user\n"
                    "\twith knowledge of system memory availability.\n"
                    "\tHowever, it's HIGHLY unlikely that you want to compute such\n"
                    "\ta large numpy array. INSTEAD, if you want all pairwise distance,\n"
                    "\tyou will likely want to set an appropriate number of cpus with TCRrep.cpus = x,\n"
                    "\tand then generate a scipy.sparse csr matrix of distances with:\n"
                    "\tTCRrep.compute_sparse_rect_distances(radius=50, chunk_size=100), leaving df and df2 arguments blank.\n"
                    "\tWhen you do this the results will be stored as TCRrep.rw_beta instead of TCRrep.pw_beta.\n"
                    "\tThis function is highly useful for comparing a smaller number of sequences against a bulk set\n"
                    "\tIn such a case, you can specify df and df2 arguments to create a non-square matrix of distances.\n"
                    "\tSee https://tcrdist3.readthedocs.io/en/latest/sparsity.html?highlight=sparse for more info.\n")
                else:
                    self.compute_distances()

            if self.archive_result:
                self.archive()
     
    def archive(self, 
                dest = None, 
                dest_tar_name = None, 
                verbose = True, 
                use_csv = False):
        """ 
        Use Zipdist2 to Make an Archive.tar.gz 

        Parameters
        ----------
        dest : str
            e.g., 'default_archive'
        dest_tar_name : str
            e.g., 'default_archive.tar.gz'
        verbose : bool
            if True, report steps in archive process
        use_csv : bool
            if True, archive will include .csv file. Useful for porting files to other applications, but creates large files.

        Example
        -------
        .. code-block:: python
            tr = TCRrep(cell_df = pd.DataFrame(), organism = "mouse")
            tr.archive(dest = "default_archive", dest_tar_name = "default_archive.tar.gz")

        Notes
        -----
        See :py:meth:`tcrdist.repertoire.rebuild`: for reubilding a TCRrep instance from 
        an TCRrep archive .tar.gz file.
        
        """
        if dest is None:
            dest = self.archive_name
        if dest_tar_name is None:
            dest_tar_name = f"{dest}.tar.gz"

        self.cell_df_index = self.cell_df.index.copy()
        self.cell_df = self.cell_df.reset_index()

        z = Zipdist2(name = dest_tar_name , target = self)
        z._save(dest = dest, dest_tar = dest_tar_name, verbose = verbose, use_csv = use_csv )
        sys.stdout.write(f"\tArchiving your TCRrep using Zipdist2 in [{dest_tar_name}]\n")
    
    def rebuild(self,
                dest = None,
                dest_tar_name = None, 
                verbose = True ):
        """ 
        Use Zipdist2 to reubild a TCRrep instance from an Archive.tar.gz

        Parameters
        ----------
        dest_tar_name : str
            e.g., 'default_archive.tar.gz'
        verbose : bool
            If True, report rebuilding process steps.

        Example
        -------
        Shows :py:meth:`tcrdist.repertoire.archive` and :py:meth:`tcrdist.repertoire.rebuild` 
        used together.

        .. code-block:: python
        
            tr = TCRrep(cell_df = pd.DataFrame(), organism = "mouse")
            tr.archive(dest = "default_archive", dest_tar_name = "default_archive.tar.gz")
            tr_new = TCRrep(cell_df = pd.DataFrame(), organism = "mouse")
            tr_new.rebuild(dest_tar_name = "default_archive.tar.gz")
        
        Notes
        -----
        See :py:meth:`tcrdist.repertoire.archive` for creating TCRrep archive file.
        """
        #tr = TCRrep(cell_df=df.iloc[0:0,:], chains=chains, organism='mouse')
        if dest is None:
            dest = self.archive_name
        if dest_tar_name is None:
            dest_tar_name = f"{dest}.tar.gz"

        z = Zipdist2(name = dest, target = self)
        z._build(dest_tar = dest_tar_name , target = self, verbose = verbose)
        
        # VALIDATION OF INPUTS
        # check that chains are valid.
        self._validate_organism()
        self._validate_chains()
        # check that  is a pd.DataFrame
        self._validate_cell_df()

        # RE INIT the REFERENCE DB see repertoire_db.py
        self.all_genes = repertoire_db.RefGeneSet(self.db_file).all_genes
       

    def compute_distances(self, df = None):
        if df is None:
            df = self.clone_df
        if 'alpha' in self.chains:
            pw_alpha  = _pws(
                df = df,
                metrics = self.metrics_a, 
                weights = self.weights_a, 
                kargs   = self.kargs_a, 
                cpu     = self.cpus, 
                store   = self.store_all_cdr)
            self._assign_distance_attributes(d = pw_alpha, chain = 'alpha')
        if 'beta' in self.chains:
            pw_beta = _pws(
                df = df, 
                metrics = self.metrics_b, 
                weights = self.weights_b, 
                kargs   = self.kargs_b, 
                cpu     = self.cpus, 
                store   = self.store_all_cdr)
            self._assign_distance_attributes(d = pw_beta, chain = 'beta')
        if 'gamma' in self.chains:
            pw_gamma = _pws(
                df = df,
                metrics = self.metrics_g, 
                weights = self.weights_g, 
                kargs   = self.kargs_g, 
                cpu     = self.cpus, 
                store   = self.store_all_cdr)
            self._assign_distance_attributes(d = pw_gamma, chain = 'gamma')
        if 'delta' in self.chains:
            pw_delta = _pws(
                df = df, 
                metrics = self.metrics_d,
                weights = self.weights_d, 
                kargs   = self.kargs_d, 
                cpu     = self.cpus, 
                store   = self.store_all_cdr)
            self._assign_distance_attributes(d = pw_delta, chain = 'delta')
    
    def compute_rect_distances(self, df = None, df2 = None, store = None):
        self._rect_distances(pw_dist_func=_pws, df = df, df2 = df2, store = store)
    
    def compute_sparse_rect_distances(self, df=None, df2=None, radius=50, chunk_size=100):
        """
        Computes sparse rectangular distance as a scipy.sparse csr matrix, only retaining distances <= radius. 
        This function is crucial when working with bulk data. Values > radius are zeros and not 
        retained saving memory. True zeros are represented as -1.
        
        NOTE if this function is called with out explicitly supplying df or df2 arguments, 
        it computes all pairwise distances between clones in the TCRrep.clone_df by default
        
        Parameters
        ----------
        df : DataFrame or None
            clone_df (will specify rows of the distance matrix)
        df2 : DataFrame or None
            clone_df (will specify columns of the distance matrix)
        radius : int
            The maximum radius, where distance information is retained the scipy.sparse csr matrix.
            All distances greater than this radius are not retained, saving memory.
        chunk_size: int
            How many rows to compute at a time. To manage memory computations are done on a chunk of rows at a time, 
            and chunks are spread out over multiple cpus. See from tcrdist.breadth import get_safe_chunk if you want
            to tune chunk size to the scale of the task.
        """
        self._rect_distances(pw_dist_func=compute_pws_sparse, df=df, df2=df2, radius=radius, chunk_size=chunk_size)
    
    def _rect_distances(self, pw_dist_func, df = None, df2 = None, **kwargs):
        if df is None:
            df = self.clone_df
        if df2 is None:
            df2 = self.df2
        if not 'store' in kwargs or kwargs['store'] is None:
            kwargs['store'] = self.store_all_cdr
        if 'alpha' in self.chains:
            pw_alpha  = pw_dist_func(
                df = df,
                df2 = df2,
                metrics = self.metrics_a, 
                weights = self.weights_a, 
                kargs   = self.kargs_a, 
                cpu     = self.cpus, 
                **kwargs)
            self._assign_distance_attributes(d = pw_alpha, chain = 'alpha',prefix = "rw")
        if 'beta' in self.chains:
            pw_beta = pw_dist_func(
                df = df,
                df2 = df2, 
                metrics = self.metrics_b, 
                weights = self.weights_b, 
                kargs   = self.kargs_b, 
                cpu     = self.cpus, 
                **kwargs)
            self._assign_distance_attributes(d = pw_beta, chain = 'beta',prefix = "rw")
        if 'gamma' in self.chains:
            pw_gamma = pw_dist_func(
                df = df,
                df2 = df2, 
                metrics = self.metrics_g, 
                weights = self.weights_g, 
                kargs   = self.kargs_g, 
                cpu     = self.cpus, 
                **kwargs)
            self._assign_distance_attributes(d = pw_gamma, chain = 'gamma', prefix = "rw")
        if 'delta' in self.chains:
            pw_delta = pw_dist_func(
                df = df,
                df2 = df2,
                metrics = self.metrics_d,
                weights = self.weights_d, 
                kargs   = self.kargs_d, 
                cpu     = self.cpus, 
                **kwargs)
            self._assign_distance_attributes(d = pw_delta, chain = 'delta', prefix = "rw")

    def _assign_distance_attributes(self, d:dict, chain:str, prefix = 'pw'):
        for k in d.keys():
            if k == 'tcrdist':
                pw_k = f"{prefix}_{chain}"
            else:
                pw_k = f"{prefix}_{k}"
            setattr(self, pw_k, d[k])

    def infer_cdrs_from_v_gene(self, chain, imgt_aligned = True):
        """
        Function taking TCR v-gene name to infer the amino amino_acid
        sequence of cdr1, cdr2, and pmhc loop regions.

        Parameters
    	----------
        chain : string
            'alpha', 'beta', 'gamma', or 'delta'
        imgt_aligned : boolean
            if True cdr1, cdr2, cdr2.5 will be returned with gaps
            and by definition will be the same length. MSH.......ET

        Returns
    	-------
        self.cell_df : pandas.core.frame.DataFrame
    	   Assigns [cdr3|cdr2|cdr1|pmhc]_[a|b|d|g]_aa columns in TCRrep.cell_df

        Examples
    	--------
        >>> testrep = TCRrep(cell_df = example_df, organism = "human", chains= ["alpha","beta"])
        >>> testrep.infer_cdrs_from_v_gene(chain = "alpha")
        >>> testrep.infer_cdrs_from_v_gene(chain = "beta")

        Notes
    	-----
        This function takes the v-gene names and infers the amino acid
        sequence of the cdr1, cdr2, and pmhc region (pmhc refers to the
        pMHC-facing loop between CDR2 and CDR3 (IMGT alignment columns 81 - 86.
      
        These sequences are based up on lookup from the dictionary here:

        self.generate_ref_genes_from_db(db_file)

        imgt_aligned : boolean
            if True cdr1, cdr2, cdr2.5 will be returned with gaps
            and by definition will be the same length.
            MSH.......ET
            FNH.......DT
            LGH.......NA

        References
        ----------

        IMGT definitions of cdr1, cdr2, and pMHC-facing can be found here
        http://www.imgt.org/IMGTScientificChart/Nomenclature/IMGT-FRCDRdefinition.html
        """

        if not imgt_aligned:
            self.imgt_aligned_status = False
            f0 = lambda v : self._map_gene_to_reference_seq2(gene = v,
                                                             cdr = 0,
                                                             organism = self.organism,
                                                             attr ='cdrs_no_gaps')
            f1 = lambda v : self._map_gene_to_reference_seq2(gene = v,
                                                             cdr = 1,
                                                             organism = self.organism,
                                                             attr ='cdrs_no_gaps')
            f2 = lambda v : self._map_gene_to_reference_seq2(gene = v,
                                                             cdr = 2,
                                                             organism = self.organism,
                                                             attr ='cdrs_no_gaps')
        else:
            self.imgt_aligned_status = True
            f0 = lambda v : self._map_gene_to_reference_seq2(gene = v,
                                                             cdr = 0,
                                                             organism = self.organism,
                                                             attr ='cdrs')
            f1 = lambda v : self._map_gene_to_reference_seq2(gene = v,
                                                             cdr = 1,
                                                             organism = self.organism,
                                                             attr ='cdrs')
            f2 = lambda v : self._map_gene_to_reference_seq2(gene = v,
                                                             cdr = 2,
                                                             organism = self.organism,
                                                             attr ='cdrs')
        if chain == "alpha":
            self.cell_df = self.cell_df.assign(cdr1_a_aa=list(map(f0, self.cell_df.v_a_gene)),
                                               cdr2_a_aa=list(map(f1, self.cell_df.v_a_gene)),
                                               pmhc_a_aa=list(map(f2, self.cell_df.v_a_gene)))
        if chain == "beta":
            self.cell_df = self.cell_df.assign(cdr1_b_aa=list(map(f0, self.cell_df.v_b_gene)),
                                               cdr2_b_aa=list(map(f1, self.cell_df.v_b_gene)),
                                               pmhc_b_aa=list(map(f2, self.cell_df.v_b_gene)))
        if chain == "gamma":
            self.cell_df = self.cell_df.assign(cdr1_g_aa=list(map(f0, self.cell_df.v_g_gene)),
                                               cdr2_g_aa=list(map(f1, self.cell_df.v_g_gene)),
                                               pmhc_g_aa=list(map(f2, self.cell_df.v_g_gene)))
        if chain == "delta":
            self.cell_df = self.cell_df.assign(cdr1_d_aa=list(map(f0, self.cell_df.v_d_gene)),
                                               cdr2_d_aa=list(map(f1, self.cell_df.v_d_gene)),
                                               pmhc_d_aa=list(map(f2, self.cell_df.v_d_gene)))
    
    def infer_index_cols(self):
        """
        Infers index columns from TCRrep.cell_df
        """
        self.index_cols = [item for item in self.cell_df.columns.to_list() if item not in ['count', 'cell_id', 'clone_id']]
        
    def deduplicate(self):
        """
        With attribute self.index_col calls _deduplicate() and assigns result to attribute self.clone_df
        """
        clone_df = _deduplicate(self.cell_df, self.index_cols)
        
        # check if any clones were lost due to missing information
        if np.sum(self.cell_df['count']) != np.sum(clone_df['count']):
            n_cells_lost = np.sum(self.cell_df['count']) - np.sum(clone_df['count'])
            n_cell = np.sum(self.cell_df['count'])
            warnings.warn(f"Not all cells/sequences could be grouped into clones."
                          f"{n_cells_lost} of {n_cell} were not captured. This occurs when "
                          "any of the values in the index columns are null or missing for a given sequence. "
                          "To see entries with missing values use: tcrdist.repertoire.TCRrep._show_incomplete()", stacklevel=2)
        
        # if no clone id column provided thetrn create one as a sequence of numbers
        if "clone_id" not in clone_df:
            N = clone_df.shape[0]
            clone_df['clone_id'] = range(1, N + 1 ,1)
        
        self.clone_df = clone_df
        return clone_df.copy()

    def show_incomplete(self):
        """
        Returns a dataframe with those cells that do not have a valid entry 
        for any one of the specified index columns
        """   
        ind = self.cell_df[self.index_cols].isnull().any(axis = 1)   
        incomplete_clones = self.cell_df.loc[ind,self.index_cols].copy()
        return incomplete_clones 

    """
        _INTERNAL functions
    """

    def _initialize_chain_specific_attributes(self):
        """
        Initialize pw object and default substitution matrix (smat) based on
        chains arguments.

        Naming of all objects have a standardized order
            region_chain_molecular_object
            (cdr3)_(a|b|d|g)_(aa|p)_(pw|smat|hmat)

        """
        if "alpha" in self.chains:
            self.cdr3_a_aa_smat = 'blosum62'
            self.cdr2_a_aa_smat = 'blosum62'
            self.cdr1_a_aa_smat = 'blosum62'
            self.pmhc_a_aa_smat = 'blosum62'
            self.index_cols.append("cdr3_a_aa")
            self.metrics_a = { 
                "cdr3_a_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_a_aa" : pw.metrics.nb_vector_tcrdist }
            self.weights_a = { 
                "cdr3_a_aa" : 3,
                "pmhc_a_aa" : 1,
                "cdr2_a_aa" : 1,
                "cdr1_a_aa" : 1}
            self.kargs_a = {
                "cdr3_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}
        
        if 'beta' in self.chains:
            self.cdr3_b_aa_smat = 'blosum62'
            self.cdr2_b_aa_smat = 'blosum62'
            self.cdr1_b_aa_smat = 'blosum62'
            self.pmhc_b_aa_smat = 'blosum62'
            self.index_cols.append("cdr3_b_aa")
            self.metrics_b = { "cdr3_b_aa" : pw.metrics.nb_vector_tcrdist,
                               "pmhc_b_aa" : pw.metrics.nb_vector_tcrdist,
                               "cdr2_b_aa" : pw.metrics.nb_vector_tcrdist,
                               "cdr1_b_aa" : pw.metrics.nb_vector_tcrdist }
            self.metrics_b = { 
                "cdr3_b_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_b_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_b_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_b_aa" : pw.metrics.nb_vector_tcrdist }
            self.weights_b = { 
                "cdr3_b_aa" : 3,
                "pmhc_b_aa" : 1,
                "cdr2_b_aa" : 1,
                "cdr1_b_aa" : 1}
            self.kargs_b = {
                "cdr3_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}

        if 'gamma' in self.chains:
            self.cdr3_g_aa_smat = 'blosum62'
            self.cdr2_g_aa_smat = 'blosum62'
            self.cdr1_g_aa_smat = 'blosum62'
            self.pmhc_g_aa_smat = 'blosum62'
            self.index_cols.append("cdr3_g_aa")
            self.metrics_g = { "cdr3_g_aa" : pw.metrics.nb_vector_tcrdist,
                               "pmhc_g_aa" : pw.metrics.nb_vector_tcrdist,
                               "cdr2_g_aa" : pw.metrics.nb_vector_tcrdist,
                               "cdr1_g_aa" : pw.metrics.nb_vector_tcrdist }
            self.metrics_g = { 
                "cdr3_g_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_g_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_g_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_g_aa" : pw.metrics.nb_vector_tcrdist }
            self.weights_g = { 
                "cdr3_g_aa" : 3,
                "pmhc_g_aa" : 1,
                "cdr2_g_aa" : 1,
                "cdr1_g_aa" : 1}
            self.kargs_g = {
                "cdr3_g_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_g_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_g_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_g_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}
      

        if 'delta' in self.chains:
            self.cdr3_d_aa_smat = 'blosum62'
            self.cdr2_d_aa_smat = 'blosum62'
            self.cdr1_d_aa_smat = 'blosum62'
            self.pmhc_d_aa_smat = 'blosum62'
            self.index_cols.append("cdr3_d_aa")
            self.metrics_d = { "cdr3_d_aa" : pw.metrics.nb_vector_tcrdist,
                               "pmhc_d_aa" : pw.metrics.nb_vector_tcrdist,
                               "cdr2_d_aa" : pw.metrics.nb_vector_tcrdist,
                               "cdr1_d_aa" : pw.metrics.nb_vector_tcrdist }
            self.metrics_d = { 
                "cdr3_d_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_d_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_d_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_d_aa" : pw.metrics.nb_vector_tcrdist }
            self.weights_d = { 
                "cdr3_d_aa" : 3,
                "pmhc_d_aa" : 1,
                "cdr2_d_aa" : 1,
                "cdr1_d_aa" : 1}
            self.kargs_d = {
                "cdr3_d_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_d_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_d_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_d_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}
      

    def synthesize_vj_matched_background(self, ts  = None, chain = "beta" , n = 100000):
        """
        tcrsampler : TCRsampler or None
       
        chain : str 
            'beta' (in future, TODO: add 'alpha')
        n : int
            number of sequences to synthesize and adjust
        TODO
        -------
        ONLY WORKS CURRENTLY FOR HUMAN BETA, VIA OLGA
        """
        if chain not in ["beta", "alpha"]:
            raise ValueError("Invalid <chain> argument.")
        if chain == "beta":
            if ts is None:
                ts = _default_sampler(organism = self.organism, chain = "beta")()
                ts = get_stratified_gene_usage_frequency(ts = ts, replace = True) 
            if self.organism == "human":
                vj_background = _synthesize_human_beta_vj_background(ts = ts, df = self.clone_df, n = n)
            elif self.organism == "mouse":
                vj_background = _synthesize_mouse_beta_vj_background(ts = ts, df = self.clone_df, n = n)
        # TODO: ADD OTHER OPTIONS
        elif chain == "alpha":
            if ts is None:
                ts = _default_sampler(organism = self.organism, chain = "alpha")()
                ts = get_stratified_gene_usage_frequency(ts = ts, replace = True) 
            if self.organism == "human":
                #raise ValueError("TODO: FUTURE VERSIONS NEED ALPHA(HUMAN)")
                vj_background = _synthesize_human_alpha_vj_background(ts = ts, df = self.clone_df, n = n)
            elif self.organism == "mouse":
                #raise ValueError("TODO: FUTURE VERSIONS NEED ALPHA(MOUSE)")
                vj_background = _synthesize_mouse_alpha_vj_background(ts = ts, df = self.clone_df, n = n)
        return vj_background   





    def _map_gene_to_reference_seq2(self,
                                    organism,
                                    gene,
                                    cdr,
                                    attr = 'cdrs_no_gaps'):
        """
        Internal function that looks up the cdr sequence (gapped or ungapped)
        from the self.all_genes library

        Parameter
        ---------
        organism : string
            mouse or human
        gene : string
            specifies the TCR gene such as 'TRAV1*01'
        cdr : int
            0 - CDR1, 1-CDR2 and 2 - CDR2.5
        attr : string
            'cdrs_no_gaps' or 'cdrs_aligned' with gaps from IMGT
        """
        try:
            aa_string = self.all_genes[organism][gene].__dict__[attr][cdr]
        except KeyError:
            aa_string = None
            warnings.warn("{} gene was not recognized in reference db no cdr seq could be inferred".format(gene), stacklevel=2)
        return(aa_string)
       
        """
        _VALIDATION functions - check validity of inputs and communicate errors to user
        """

    def _validate_db_file(self):
        """
        Issues warning if invalid organism is passed to TCRrep __init__
        """
        if self.db_file not in ['alphabeta_gammadelta_db.tsv','alphabeta_db.tsv','gammadelta_db.tsv']:
            warnings.warn("db_file must be 'alphabeta_gammadelta_db.tsv' or 'alphabeta_db.tsv' or 'gammadelta_db.tsv' unless you have built tcrdist3 from scratch", stacklevel=2)
    
    def _validate_organism(self):
        """
        Raises 
        ------
        ValueError if invalid organism is passed to TCRrep __init__
        """
        if self.organism not in ["human", "mouse", "rhesus"]:
            raise ValueError("organism must be 'mouse', 'human', or 'rhesus'")
    
    def _validate_chains(self):
        """
        Raises
        ------ 
        ValueError if invalid chains are passed to TCRrep __init__
        """
        check_chains_arg = ['alpha', 'beta', "gamma", "delta"]
        if len([c for c in self.chains if c not in check_chains_arg]) > 0:
            raise ValueError('TCRrep chains arg can be one or more of the '
                                'following {} case-sensitive'.format(check_chains_arg))
    def _validate_cell_df(self):
        """
        Issues warning if TCRrep.cell_df not properly formatted.
        """
        if not isinstance(self.cell_df, pd.DataFrame):
            warnings.warn('TCRrep cell_df argument must be pandas.DataFrame unless you are providing a clone_df directly\n', stacklevel=2)

        else:
            cell_df_columns = self.cell_df.columns.to_list()
            
            if "count" not in cell_df_columns:
                warnings.warn("cell_df needs a counts column to track clonal number of frequency\n", stacklevel=2)
                warnings.warn("No 'count' column provided; count column set to 1")
                self.cell_df['count'] = 1
            if "alpha" in self.chains:
                if not "cdr3_a_aa" in cell_df_columns:
                    warnings.warn("cell_df needs a column called 'cdr3_a_aa' to track the CDR3 amino acid sequence\n", stacklevel=2)
                if not "v_a_gene" in cell_df_columns:
                    warnings.warn("cell_df needs a column called 'v_a_gene' for default functions\n", stacklevel=2)
            if "beta" in self.chains:
                if not "cdr3_b_aa" in cell_df_columns:
                    warnings.warn("cell_df needs a column called 'cdr3_b_aa' to track the CDR3 amino acid sequence\n", stacklevel=2)
                if not "v_b_gene" in cell_df_columns:
                    warnings.warn("cell_df needs a column called 'v_b_gene' for default functions\n", stacklevel=2)
            if "gamma" in self.chains:
                if not "cdr3_g_aa" in cell_df_columns:
                    warnings.warn("cell_df needs a column called 'cdr3_g_aa' to track the CDR3 amino acid sequence\n", stacklevel=2)
                if not "v_g_gene" in cell_df_columns:
                    warnings.warn("cell_df needs a column called 'v_g_gene' for default functions\n", stacklevel=2)
            if "delta" in self.chains:
                if not "cdr3_d_aa" in cell_df_columns:
                    warnings.warn("cell_df needs a column called 'cdr3_d_aa' to track the CDR3 amino acid sequence\n", stacklevel=2)
                if not "v_d_gene" in cell_df_columns:
                    warnings.warn("cell_df needs a column called 'v_d_gene' for default functions\n", stacklevel=2)

    def _validate_imgt_aligned(self):
        """
        Raises
        ------
        ValueError if TCRrep.imgt_aligned is not a boolean
        """
        if not isinstance(self.imgt_aligned, bool):
            raise ValueError('TCRrep imgt_aligned argument must be a boolean')

    # def _validate_cdr_sequences():
    #     pass
    #     # TODO: Avoid trouble by wanring the user that they have non valid characters 

    
def _deduplicate(cell_df, index_cols):
    """
    Use index_cols to group by and group identical entries. The input DataFrame
    must have a column 'count'.
    """
    clones = cell_df.groupby(index_cols)['count'].agg('sum').reset_index()
    return clones
